#!/usr/bin/python

# Copyright (c) 2013, Johan Wiren <johan.wiren.se@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: gem
short_description: Manage Ruby gems
description:
  - Manage installation and uninstallation of Ruby gems.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    type: str
    description:
      - The name of the gem to be managed.
    required: true
  state:
    type: str
    description:
      - The desired state of the gem. V(latest) ensures that the latest version is installed.
    choices: [present, absent, latest]
    default: present
  gem_source:
    type: path
    description:
      - The path to a local gem used as installation source.
  include_dependencies:
    description:
      - Whether to include dependencies or not.
    type: bool
    default: true
  repository:
    type: str
    description:
      - The repository from which the gem is installed.
    aliases: [source]
  user_install:
    description:
      - Install gem in user's local gems cache or for all users.
    type: bool
    default: true
  executable:
    type: path
    description:
      - Override the path to the gem executable.
  install_dir:
    type: path
    description:
      - Install the gems into a specific directory. These gems are independent from the global installed ones. Specifying
        this requires user_install to be false.
  bindir:
    type: path
    description:
      - Install executables into a specific directory.
    version_added: 3.3.0
  norc:
    type: bool
    default: true
    description:
      - Avoid loading any C(.gemrc) file. Ignored for RubyGems prior to 2.5.2.
      - The default changed from V(false) to V(true) in community.general 6.0.0.
    version_added: 3.3.0
  env_shebang:
    description:
      - Rewrite the shebang line on installed scripts to use /usr/bin/env.
    default: false
    type: bool
  version:
    type: str
    description:
      - Version of the gem to be installed/removed.
  pre_release:
    description:
      - Allow installation of pre-release versions of the gem.
    default: false
    type: bool
  include_doc:
    description:
      - Install with or without docs.
    default: false
    type: bool
  build_flags:
    type: str
    description:
      - Allow adding build flags for gem compilation.
  force:
    description:
      - Force gem to (un-)install, bypassing dependency checks.
    default: false
    type: bool
  override_platform_install_dir:
    description:
      - Resolve the user gem installation directory via C(gem environment) and pass it explicitly
        as C(--install-dir) to both C(gem install) and C(gem uninstall), instead of using C(--user-install).
      - This is needed on distributions (such as Fedora) where a platform-specific C(operating_system.rb)
        injects C(--install-dir) as a default for all gem commands, which conflicts with C(--user-install)
        and causes C(gem uninstall) to search the wrong directory.
      - Cannot be combined with O(user_install=false) or O(install_dir).
    default: false
    type: bool
    version_added: 13.0.0
author:
  - "Ansible Core Team"
  - "Johan Wiren (@johanwiren)"
"""

EXAMPLES = r"""
- name: Install version 1.0 of vagrant
  community.general.gem:
    name: vagrant
    version: 1.0
    state: present

- name: Install latest available version of rake
  community.general.gem:
    name: rake
    state: latest

- name: Install rake version 1.0 from a local gem on disk
  community.general.gem:
    name: rake
    gem_source: /path/to/gems/rake-1.0.gem
    state: present
"""

import os
import re

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils import _cmd_runner_fmt as fmt
from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner

RE_VERSION = re.compile(r"^(\d+)\.(\d+)\.(\d+)")
RE_INSTALLED = re.compile(r"\S+\s+\((?:default: )?(.+)\)")


def get_rubygems_path(module: AnsibleModule) -> list[str]:
    if module.params["executable"]:
        return module.params["executable"].split()
    return [module.get_bin_path("gem", True)]


def get_user_install_dir(module: AnsibleModule) -> str | None:
    cmd = get_rubygems_path(module)
    rc, out, err = module.run_command(cmd + ["environment"], check_rc=True)
    for line in out.splitlines():
        match = re.search(r"USER INSTALLATION DIRECTORY:\s*(.+)", line)
        if match:
            return match.group(1).strip()
    return None


def get_rubygems_version(module: AnsibleModule) -> tuple[int, ...] | None:
    cmd = get_rubygems_path(module) + ["--version"]
    rc, out, err = module.run_command(cmd, check_rc=True)
    match = RE_VERSION.match(out)
    if not match:
        return None
    return tuple(int(x) for x in match.groups())


def make_runner(module: AnsibleModule, ver: tuple[int, ...] | None) -> CmdRunner:
    command = get_rubygems_path(module)

    environ_update = {}
    if module.params["install_dir"]:
        environ_update["GEM_HOME"] = module.params["install_dir"]

    if ver and ver < (2, 0, 0):
        include_dependencies_fmt = fmt.as_bool("--include-dependencies", "--ignore-dependencies")
        include_doc_fmt = fmt.as_bool_not(["--no-rdoc", "--no-ri"])
    else:
        include_dependencies_fmt = fmt.as_bool_not("--ignore-dependencies")
        include_doc_fmt = fmt.as_bool_not("--no-document")

    norc_fmt = fmt.as_bool("--norc") if (ver and ver >= (2, 5, 2)) else fmt.as_bool([])

    return CmdRunner(
        module,
        command=command,
        environ_update=environ_update,
        check_rc=True,
        arg_formats=dict(
            _list_subcmd=fmt.as_fixed("list"),
            _install_subcmd=fmt.as_fixed("install"),
            _uninstall_subcmd=fmt.as_fixed("uninstall"),
            norc=norc_fmt,
            _remote_flag=fmt.as_fixed("--remote"),
            repository=fmt.as_opt_val("--source"),
            _name_pattern=fmt.as_func(lambda v: [f"^{v}$"]),
            version=fmt.as_opt_val("--version"),
            include_dependencies=include_dependencies_fmt,
            user_install=fmt.as_bool("--user-install", "--no-user-install"),
            install_dir=fmt.as_opt_val("--install-dir"),
            bindir=fmt.as_opt_val("--bindir"),
            pre_release=fmt.as_bool("--pre"),
            include_doc=include_doc_fmt,
            env_shebang=fmt.as_bool("--env-shebang"),
            gem_source=fmt.as_list(),
            build_flags=fmt.as_opt_val("--"),
            force=fmt.as_bool("--force"),
            _uninstall_version=fmt.as_func(lambda v: ["--version", v] if v else ["--all"], ignore_none=False),
            _executable_flag=fmt.as_fixed("--executable"),
            name=fmt.as_list(),
        ),
    )


def get_installed_versions(runner: CmdRunner, remote: bool = False) -> list[str]:
    name = runner.module.params["name"]
    if remote:
        args_order = ["_list_subcmd", "norc", "_remote_flag", "repository", "_name_pattern"]
    else:
        args_order = ["_list_subcmd", "norc", "_name_pattern"]
    with runner(args_order) as ctx:
        rc, out, err = ctx.run(_name_pattern=name)
    installed_versions = []
    for line in out.splitlines():
        match = RE_INSTALLED.match(line)
        if match:
            versions = match.group(1)
            for version in versions.split(", "):
                installed_versions.append(version.split()[0])
    return installed_versions


def exists(runner: CmdRunner) -> bool:
    module = runner.module
    if module.params["state"] == "latest":
        remoteversions = get_installed_versions(runner, remote=True)
        if remoteversions:
            module.params["version"] = remoteversions[0]
    installed_versions = get_installed_versions(runner)
    if module.params["version"]:
        return module.params["version"] in installed_versions
    return bool(installed_versions)


def install(runner: CmdRunner, user_dir: str | None = None) -> None:
    args_order = [
        "_install_subcmd",
        "norc",
        "version",
        "repository",
        "include_dependencies",
        "user_install",
        "install_dir",
        "bindir",
        "pre_release",
        "include_doc",
        "env_shebang",
        "gem_source",
        "build_flags",
        "force",
    ]
    with runner(args_order, check_mode_skip=True) as ctx:
        if user_dir:
            bindir = runner.module.params["bindir"] or os.path.join(user_dir, "bin")
            ctx.run(user_install=False, install_dir=user_dir, bindir=bindir)
        else:
            ctx.run()


def uninstall(runner: CmdRunner, user_dir: str | None = None) -> tuple[int, str, str] | None:
    args_order = [
        "_uninstall_subcmd",
        "norc",
        "install_dir",
        "bindir",
        "_uninstall_version",
        "_executable_flag",
        "force",
        "name",
    ]
    with runner(args_order, check_mode_skip=True) as ctx:
        kwargs = {"_uninstall_version": runner.module.params["version"]}
        if user_dir:
            kwargs["install_dir"] = user_dir
        return ctx.run(**kwargs)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            executable=dict(type="path"),
            gem_source=dict(type="path"),
            include_dependencies=dict(default=True, type="bool"),
            name=dict(required=True, type="str"),
            repository=dict(aliases=["source"], type="str"),
            state=dict(default="present", choices=["present", "absent", "latest"], type="str"),
            user_install=dict(default=True, type="bool"),
            install_dir=dict(type="path"),
            bindir=dict(type="path"),
            norc=dict(type="bool", default=True),
            pre_release=dict(default=False, type="bool"),
            include_doc=dict(default=False, type="bool"),
            env_shebang=dict(default=False, type="bool"),
            version=dict(type="str"),
            build_flags=dict(type="str"),
            force=dict(default=False, type="bool"),
            override_platform_install_dir=dict(default=False, type="bool"),
        ),
        supports_check_mode=True,
        mutually_exclusive=[["gem_source", "repository"], ["gem_source", "version"]],
    )

    if module.params["version"] and module.params["state"] == "latest":
        module.fail_json(msg="Cannot specify version when state=latest")
    if module.params["gem_source"] and module.params["state"] == "latest":
        module.fail_json(msg="Cannot maintain state=latest when installing from local source")
    if module.params["user_install"] and module.params["install_dir"]:
        module.fail_json(msg="install_dir requires user_install=false")
    if module.params["override_platform_install_dir"]:
        if not module.params["user_install"]:
            module.fail_json(msg="override_platform_install_dir requires user_install=true")
        if module.params["install_dir"]:
            module.fail_json(msg="override_platform_install_dir cannot be combined with install_dir")

    if not module.params["gem_source"]:
        module.params["gem_source"] = module.params["name"]

    ver = get_rubygems_version(module)

    user_dir = None
    if module.params["override_platform_install_dir"]:
        user_dir = get_user_install_dir(module)

    runner = make_runner(module, ver)

    changed = False

    if module.params["state"] in ["present", "latest"]:
        if not exists(runner):
            install(runner, user_dir)
            changed = True
    elif module.params["state"] == "absent":
        if exists(runner):
            command_output = uninstall(runner, user_dir)
            if command_output is not None and exists(runner):
                rc, out, err = command_output
                module.fail_json(
                    msg=(
                        f"Failed to uninstall gem '{module.params['name']}': it is still present after 'gem uninstall'. "
                        "This usually happens with default or system gems provided by the OS, "
                        "which cannot be removed with the gem command."
                    ),
                    rc=rc,
                    stdout=out,
                    stderr=err,
                )
            else:
                changed = True
    result = {}
    result["name"] = module.params["name"]
    result["state"] = module.params["state"]
    if module.params["version"]:
        result["version"] = module.params["version"]
    result["changed"] = changed

    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2019, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t

PEM_START = "-----BEGIN "
PEM_END_START = "-----END "
PEM_END = "-----"
PKCS8_PRIVATEKEY_NAMES = ("PRIVATE KEY", "ENCRYPTED PRIVATE KEY")
PKCS1_PRIVATEKEY_SUFFIX = " PRIVATE KEY"


def identify_pem_format(content: bytes, *, encoding: str = "utf-8") -> bool:
    """Given the contents of a binary file, tests whether this could be a PEM file."""
    try:
        first_pem = extract_first_pem(content.decode(encoding))
        if first_pem is None:
            return False
        lines = first_pem.splitlines(False)
        if (
            lines[0].startswith(PEM_START)
            and lines[0].endswith(PEM_END)
            and len(lines[0]) > len(PEM_START) + len(PEM_END)
        ):
            return True
    except UnicodeDecodeError:
        pass
    return False


def identify_private_key_format(
    content: bytes, *, encoding: str = "utf-8"
) -> t.Literal["raw", "pkcs1", "pkcs8", "unknown-pem"]:
    """Given the contents of a private key file, identifies its format."""
    # See https://github.com/openssl/openssl/blob/master/crypto/pem/pem_pkey.c#L40-L85
    # (PEM_read_bio_PrivateKey)
    # and https://github.com/openssl/openssl/blob/master/include/openssl/pem.h#L46-L47
    # (PEM_STRING_PKCS8, PEM_STRING_PKCS8INF)
    try:
        first_pem = extract_first_pem(content.decode(encoding))
        if first_pem is None:
            return "raw"
        lines = first_pem.splitlines(False)
        if (
            lines[0].startswith(PEM_START)
            and lines[0].endswith(PEM_END)
            and len(lines[0]) > len(PEM_START) + len(PEM_END)
        ):
            name = lines[0][len(PEM_START) : -len(PEM_END)]
            if name in PKCS8_PRIVATEKEY_NAMES:
                return "pkcs8"
            if len(name) > len(PKCS1_PRIVATEKEY_SUFFIX) and name.endswith(
                PKCS1_PRIVATEKEY_SUFFIX
            ):
                return "pkcs1"
            return "unknown-pem"
    except UnicodeDecodeError:
        pass
    return "raw"


def split_pem_list(text: str, *, keep_inbetween: bool = False) -> list[str]:
    """
    Split concatenated PEM objects into a list of strings, where each is one PEM object.
    """
    result = []
    current: list[str] | None = [] if keep_inbetween else None
    for line in text.splitlines(True):
        if line.strip():
            if not keep_inbetween and line.startswith("-----BEGIN "):
                current = []
            if current is not None:
                current.append(line)
                if line.startswith("-----END "):
                    result.append("".join(current))
                    current = [] if keep_inbetween else None
    return result


def extract_first_pem(text: str) -> str | None:
    """
    Given one PEM or multiple concatenated PEM objects, return only the first one, or None if there is none.
    """
    all_pems = split_pem_list(text)
    if not all_pems:
        return None
    return all_pems[0]


def _extract_type(line: str, *, start: str = PEM_START) -> str | None:
    if not line.startswith(start):
        return None
    if not line.endswith(PEM_END):
        return None
    return line[len(start) : -len(PEM_END)]


def extract_pem(content: str, *, strict: bool = False) -> tuple[str, str]:
    lines = content.splitlines()
    if len(lines) < 3:
        raise ValueError(f"PEM must have at least 3 lines, have only {len(lines)}")
    header_type = _extract_type(lines[0])
    if header_type is None:
        raise ValueError(
            f"First line is not of format {PEM_START}...{PEM_END}: {lines[0]!r}"
        )
    footer_type = _extract_type(lines[-1], start=PEM_END_START)
    if strict:
        if header_type != footer_type:
            raise ValueError(
                f"Header type ({header_type}) is different from footer type ({footer_type})"
            )
        for idx, line in enumerate(lines[1:-2]):
            if len(line) != 64:
                raise ValueError(f"Line {idx} has length {len(line)} instead of 64")
        if not (0 < len(lines[-2]) <= 64):
            raise ValueError(
                f"Last line has length {len(lines[-2])}, should be in (0, 64]"
            )
    return header_type, "".join(lines[1:-1])


__all__ = (
    "PEM_END",
    "PEM_END_START",
    "PEM_START",
    "PKCS1_PRIVATEKEY_SUFFIX",
    "PKCS8_PRIVATEKEY_NAMES",
    "extract_first_pem",
    "extract_pem",
    "identify_pem_format",
    "identify_private_key_format",
    "split_pem_list",
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Sabari Jaganathan (@sajagana) <sajagana@cisco.com>

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: ndo_route_map_policy_route_control
version_added: "2.12.0"
short_description: Manage Route Map Policy for Route Control on Cisco Nexus Dashboard Orchestrator (NDO).
description:
- Manage Route Map Policy for Route Control on Cisco Nexus Dashboard Orchestrator (NDO).
- This module is only supported on ND v4.1 and later.
author:
- Sabari Jaganathan (@sajagana)
options:
  template:
    description:
    - The name of the template.
    - The template must be a tenant template.
    - This parameter or O(template_id) is required.
    type: str
  template_id:
    description:
    - The ID of the tenant template.
    - This parameter or O(template) is required.
    type: str
  name:
    description:
    - The name of the Route Map Policy for Route Control.
    type: str
  uuid:
    description:
    - The UUID of the Route Map Policy for Route Control.
    - This parameter is required when the O(name) needs to be updated.
    type: str
  description:
    description:
    - The description of the Route Map Policy for Route Control.
    type: str
  match_type:
    description:
    - The match type of the Route Map Policy for Route Control.
    - Defaults to C(routing_policy_only) when unset during creation.
    type: str
    choices: [ prefix_and_routing_policy, routing_policy_only ]
  state:
    description:
    - Use C(absent) for removing.
    - Use C(query) for listing an object or multiple objects.
    - Use C(present) for creating or updating.
    type: str
    choices: [ absent, query, present ]
    default: query
notes:
- The O(template) must exist before using this module in your playbook.
  Use M(cisco.mso.ndo_template) to create the Tenant template.
seealso:
- module: cisco.mso.ndo_template
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add Route Map Policy for Route Control
  cisco.mso.ndo_route_map_policy_route_control:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    name: route_map_policy_1
    state: present
  register: add_route_map_policy

- name: Update Route Map Policy for Route Control name using UUID
  cisco.mso.ndo_route_map_policy_route_control:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    name: route_map_policy_1_updated
    uuid: "{{ add_route_map_policy.current.uuid }}"
    match_type: prefix_and_routing_policy
    state: present

- name: Query Route Map Policy for Route Control using name
  cisco.mso.ndo_route_map_policy_route_control:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    name: route_map_policy_1_updated
    state: query
  register: query_one_with_name

- name: Query Route Map Policy for Route Control using UUID
  cisco.mso.ndo_route_map_policy_route_control:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    uuid: "{{ add_route_map_policy.current.uuid }}"
    state: query
  register: query_one_with_uuid

- name: Query all Route Map Policy for Route Controls
  cisco.mso.ndo_route_map_policy_route_control:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    state: query
  register: query_all

- name: Remove Route Map Policy for Route Control using name
  cisco.mso.ndo_route_map_policy_route_control:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    name: route_map_policy_1_updated
    state: absent

- name: Remove Route Map Policy for Route Control using UUID
  cisco.mso.ndo_route_map_policy_route_control:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    uuid: "{{ add_route_map_policy.current.uuid }}"
    state: absent
"""

RETURN = r"""
"""


import copy
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.constants import ROUTE_MAP_POLICY_MATCH_TYPE
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.templates import MSOTemplates
from ansible_collections.cisco.mso.plugins.module_utils.utils import append_update_ops_data


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        template=dict(type="str"),
        template_id=dict(type="str"),
        name=dict(type="str"),
        uuid=dict(type="str"),
        description=dict(type="str"),
        match_type=dict(type="str", choices=list(ROUTE_MAP_POLICY_MATCH_TYPE)),
        state=dict(type="str", default="query", choices=["absent", "query", "present"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["name", "uuid"], True],
            ["state", "present", ["name", "uuid"], True],
        ],
        required_one_of=[
            ["template", "template_id"],
        ],
    )

    mso = MSOModule(module)
    mso_templates = MSOTemplates(mso)

    template_name = mso.params.get("template")
    template_id = mso.params.get("template_id")
    name = mso.params.get("name")
    uuid = mso.params.get("uuid")
    description = mso.params.get("description")
    match_type = ROUTE_MAP_POLICY_MATCH_TYPE.get(mso.params.get("match_type"))
    state = mso.params.get("state")

    ops = []
    match = None
    path = None

    mso_template = mso_templates.get_template("tenant", template_name, template_id)
    mso_template.validate_template("tenantPolicy")

    match = mso_template.get_route_map_policy(uuid, name)
    if (uuid or name) and match:  # Query a specific object
        mso.existing = mso.previous = copy.deepcopy(mso_template.update_config_with_template_and_references(match.details))
    elif match:  # Query all objects
        mso.existing = [mso_template.update_config_with_template_and_references(obj) for obj in match]

    if state != "query":
        path = "/tenantPolicyTemplate/template/routeMapPolicies/{0}".format(match.index if match else "-")

    if state == "present":
        mso_values = {
            "name": name,
            "description": description,
            "routeMapType": match_type,
        }
        if match:
            append_update_ops_data(ops, match.details, path, mso_values)
            mso.sanitize(mso_values, collate=True)
        else:
            mso.sanitize(mso_values)
            ops.append(dict(op="add", path=path, value=mso.sent))

    elif state == "absent" and match:
        ops.append(dict(op="remove", path=path))

    if mso.proposed:
        mso.proposed = copy.deepcopy(mso.proposed)
        mso_template.update_config_with_template_and_references(mso.proposed)

    if not module.check_mode and ops:
        response = mso.request(mso_template.template_path, method="PATCH", data=ops)
        match = mso_template.get_route_map_policy(uuid, name, template_object=response)
        if match:
            mso.existing = mso_template.update_config_with_template_and_references(match.details)  # When the state is present
        else:
            mso.existing = {}  # When the state is absent
    elif module.check_mode and state != "query":  # When the state is present/absent with check mode
        mso.existing = mso.proposed if state == "present" else {}

    mso.exit_json()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2025 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the
# ansible.content_builder.
#
# Manually editing this file is not advised.
#
# To update the argspec make the desired changes
# in the documentation in the module file and re-run
# ansible.content_builder commenting out
# the path to external 'docstring' in build.yaml.
#
##############################################

"""
The arg spec for the nxos_l3_interfaces module
"""


class L3_interfacesArgs(object):  # pylint: disable=R0903
    """The arg spec for the nxos_l3_interfaces module"""

    argument_spec = {
        "config": {
            "type": "list",
            "elements": "dict",
            "options": {
                "name": {"type": "str", "required": True},
                "mac_address": {"type": "str"},
                "bandwidth": {
                    "type": "dict",
                    "options": {
                        "kilobits": {"type": "int"},
                        "inherit": {"type": "bool"},
                    },
                },
                "dot1q": {"type": "int"},
                "evpn_multisite_tracking": {
                    "type": "str",
                    "choices": ["fabric-tracking", "dci-tracking"],
                },
                "ipv6_redirects": {"type": "bool"},
                "ipv6_unreachables": {"type": "bool"},
                "redirects": {"type": "bool"},
                "unreachables": {"type": "bool"},
                "proxy_arp": {"type": "bool"},
                "port_unreachable": {"type": "bool"},
                "verify": {
                    "type": "dict",
                    "options": {
                        "unicast": {
                            "type": "dict",
                            "options": {
                                "source": {
                                    "type": "dict",
                                    "options": {
                                        "reachable_via": {
                                            "type": "dict",
                                            "options": {
                                                "mode": {"type": "str"},
                                                "allow_default": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                            },
                        },
                    },
                },
                "ipv6_verify": {
                    "type": "dict",
                    "options": {
                        "unicast": {
                            "type": "dict",
                            "options": {
                                "source": {
                                    "type": "dict",
                                    "options": {
                                        "reachable_via": {
                                            "type": "dict",
                                            "options": {
                                                "mode": {"type": "str"},
                                                "allow_default": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                            },
                        },
                    },
                },
                "dhcp": {
                    "type": "dict",
                    "options": {
                        "ipv4": {
                            "type": "dict",
                            "options": {
                                "option82": {
                                    "type": "dict",
                                    "options": {
                                        "suboption": {
                                            "type": "dict",
                                            "options": {
                                                "circuit_id": {"type": "str"},
                                            },
                                        },
                                    },
                                },
                                "smart_relay": {"type": "bool"},
                                "relay": {
                                    "type": "dict",
                                    "options": {
                                        "information": {
                                            "type": "dict",
                                            "options": {
                                                "trusted": {"type": "bool"},
                                            },
                                        },
                                        "subnet_selection": {
                                            "type": "dict",
                                            "options": {
                                                "subnet_ip": {"type": "str"},
                                            },
                                        },
                                        "source_interface": {
                                            "type": "dict",
                                            "options": {
                                                "interface_type": {"type": "str"},
                                                "interface_id": {"type": "str"},
                                            },
                                        },
                                        "address": {
                                            "type": "list",
                                            "elements": "dict",
                                            "options": {
                                                "relay_ip": {"type": "str"},
                                                "vrf_name": {"type": "str"},
                                            },
                                        },
                                    },
                                },
                            },
                        },
                        "ipv6": {
                            "type": "dict",
                            "options": {
                                "smart_relay": {"type": "bool"},
                                "relay": {
                                    "type": "dict",
                                    "options": {
                                        "source_interface": {
                                            "type": "dict",
                                            "options": {
                                                "interface_type": {"type": "str"},
                                                "interface_id": {"type": "str"},
                                            },
                                        },
                                        "address": {
                                            "type": "list",
                                            "elements": "dict",
                                            "options": {
                                                "relay_ip": {"type": "str"},
                                                "vrf_name": {"type": "str"},
                                                "interface_type": {"type": "str"},
                                                "interface_id": {"type": "str"},
                                            },
                                        },
                                    },
                                },
                            },
                        },
                    },
                },
                "ipv4": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "address": {"type": "str"},
                        "tag": {"type": "int"},
                        "secondary": {"type": "bool"},
                        "ip_network_mask": {"type": "str"},
                        "route_preference": {"type": "int"},
                    },
                },
                "ipv6": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "address": {"type": "str"},
                        "tag": {"type": "int"},
                        "aggregate_prefix_length": {"type": "int"},
                        "anycast": {"type": "bool"},
                        "eui64": {"type": "bool"},
                        "route_preference": {"type": "int"},
                        "use_bia": {"type": "bool"},
                    },
                },
            },
        },
        "running_config": {"type": "str"},
        "state": {
            "choices": [
                "merged",
                "replaced",
                "overridden",
                "deleted",
                "rendered",
                "gathered",
                "parsed",
            ],
            "default": "merged",
            "type": "str",
        },
    }  # pylint: disable=C0301



# Copyright (c) 2016 Michael Gruener <michael.gruener@chaosmoon.net>
# Copyright (c) 2021 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import copy
import datetime
import json
import locale
import time
import typing as t

from ansible.module_utils.basic import missing_required_lib
from ansible.module_utils.common.text.converters import to_bytes
from ansible.module_utils.urls import fetch_url

from ansible_collections.community.crypto.plugins.module_utils._acme.backend_cryptography import (
    CRYPTOGRAPHY_ERROR,
    CRYPTOGRAPHY_MINIMAL_VERSION,
    CRYPTOGRAPHY_VERSION,
    HAS_CURRENT_CRYPTOGRAPHY,
    CryptographyBackend,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.backend_openssl_cli import (
    OpenSSLCLIBackend,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    ACMEProtocolException,
    KeyParsingError,
    ModuleFailException,
    NetworkException,
    format_http_status,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.utils import (
    compute_cert_id,
    nopad_b64,
    parse_retry_after,
)
from ansible_collections.community.crypto.plugins.module_utils._argspec import (
    ArgumentSpec,
)
from ansible_collections.community.crypto.plugins.module_utils._time import (
    get_now_datetime,
)

if t.TYPE_CHECKING:
    import http.client  # pragma: no cover
    import os  # pragma: no cover
    import urllib.error  # pragma: no cover

    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover

    from ansible_collections.community.crypto.plugins.module_utils._acme.backends import (  # pragma: no cover
        CertificateInformation,
        CryptoBackend,
    )


# -1 usually means connection problems
RETRY_STATUS_CODES = (-1, 408, 429, 502, 503, 504)

RETRY_COUNT = 20


def _decode_retry(
    *,
    module: AnsibleModule,
    response: urllib.error.HTTPError | http.client.HTTPResponse | None,
    info: dict[str, t.Any],
    retry_count: int,
) -> bool:
    if info["status"] not in RETRY_STATUS_CODES:
        return False

    if retry_count >= RETRY_COUNT:
        raise ACMEProtocolException(
            module=module,
            msg=f"Giving up after {RETRY_COUNT} retries",
            info=info,
            response=response,
        )

    # 429 and 503 should have a Retry-After header (https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Retry-After)
    now = get_now_datetime(with_timezone=True)
    try:
        then = parse_retry_after(
            info.get("retry-after", "10"), relative_with_timezone=True, now=now
        )
        retry_after = (then - now).total_seconds()
        retry_after = min(max(1, retry_after), 60)
    except (TypeError, ValueError):
        retry_after = 10
    module.log(
        f"Retrieved a {format_http_status(info['status'])} HTTP status on {info['url']}, retrying in {retry_after} seconds"
    )

    time.sleep(retry_after)
    return True


def _assert_fetch_url_success(
    *,
    module: AnsibleModule,
    response: urllib.error.HTTPError | http.client.HTTPResponse | None,
    info: dict[str, t.Any],
    allow_redirect: bool = False,
    allow_client_error: bool = True,
    allow_server_error: bool = True,
) -> None:
    if info["status"] < 0:
        raise NetworkException(msg=f"Failure downloading {info['url']}, {info['msg']}")

    if (
        (300 <= info["status"] < 400 and not allow_redirect)
        or (400 <= info["status"] < 500 and not allow_client_error)
        or (info["status"] >= 500 and not allow_server_error)
    ):
        raise ACMEProtocolException(module=module, info=info, response=response)


def _is_failed(
    *, info: dict[str, t.Any], expected_status_codes: t.Iterable[int] | None = None
) -> bool:
    if info["status"] < 200 or info["status"] >= 400:
        return True
    return bool(
        expected_status_codes is not None
        and info["status"] not in expected_status_codes
    )


class ACMEDirectory:
    """
    The ACME server directory. Gives access to the available resources,
    and allows to obtain a Replay-Nonce. The acme_directory URL
    needs to support unauthenticated GET requests; ACME endpoints
    requiring authentication are not supported.
    https://tools.ietf.org/html/rfc8555#section-7.1.1
    """

    def __init__(self, *, module: AnsibleModule, client: ACMEClient) -> None:
        self.module = module
        self.directory_root = module.params["acme_directory"]
        self.version = module.params["acme_version"]

        directory, info = client.get_request(self.directory_root, get_only=True)
        if not isinstance(directory, dict):
            raise ACMEProtocolException(
                module=module,
                msg=f"ACME directory is not a dictionary, but {type(directory)}",
                info=info,
                content_json=directory,
            )
        self.directory = directory

        self.request_timeout = module.params["request_timeout"]

        # Check whether self.version matches what we expect
        if self.version == 2:
            for key in ("newNonce", "newAccount", "newOrder"):
                if key not in self.directory:
                    raise ModuleFailException(
                        "ACME directory does not seem to follow protocol ACME v2"
                    )
            # Make sure that 'meta' is always available
            if "meta" not in self.directory:
                self.directory["meta"] = {}

    def __getitem__(self, key: str) -> t.Any:
        return self.directory[key]

    def __contains__(self, key: str) -> bool:
        return key in self.directory

    def get(self, key: str, default_value: t.Any = None) -> t.Any:
        return self.directory.get(key, default_value)

    def get_nonce(self, resource: str | None = None) -> str:
        url = self.directory["newNonce"]
        if resource is not None:
            url = resource
        retry_count = 0
        while True:
            response, info = fetch_url(
                self.module, url, method="HEAD", timeout=self.request_timeout
            )
            if _decode_retry(
                module=self.module,
                response=response,
                info=info,
                retry_count=retry_count,
            ):
                retry_count += 1
                continue
            if info["status"] not in (200, 204):
                raise NetworkException(
                    f"Failed to get replay-nonce, got status {format_http_status(info['status'])}"
                )
            if "replay-nonce" in info:
                return info["replay-nonce"]
            self.module.log(
                f"HEAD to {url} did return status {format_http_status(info['status'])}, but no replay-nonce header!"
            )
            if retry_count >= 5:
                raise ACMEProtocolException(
                    module=self.module,
                    msg="Was not able to obtain nonce, giving up after 5 retries",
                    info=info,
                    response=response,
                )
            retry_count += 1

    def has_renewal_info_endpoint(self) -> bool:
        return "renewalInfo" in self.directory


class ACMEClient:
    """
    ACME client object. Handles the authorized communication with the
    ACME server.
    """

    def __init__(self, *, module: AnsibleModule, backend: CryptoBackend) -> None:
        # Set to true to enable logging of all signed requests
        self._debug = False

        self.module = module
        self.backend = backend
        self.version = module.params["acme_version"]
        # account_key path and content are mutually exclusive
        self.account_key_file = module.params.get("account_key_src")
        self.account_key_content = module.params.get("account_key_content")
        self.account_key_passphrase = module.params.get("account_key_passphrase")

        # Grab account URI from module parameters.
        # Make sure empty string is treated as None.
        self.account_uri = module.params.get("account_uri") or None

        self.request_timeout = module.params["request_timeout"]

        self.account_key_data = None
        self.account_jwk = None
        self.account_jws_header = None
        if self.account_key_file is not None or self.account_key_content is not None:
            try:
                self.account_key_data = self.parse_key(
                    key_file=self.account_key_file,
                    key_content=self.account_key_content,
                    passphrase=self.account_key_passphrase,
                )
            except KeyParsingError as e:
                raise ModuleFailException(
                    f"Error while parsing account key: {e.msg}"
                ) from e
            self.account_jwk = self.account_key_data["jwk"]
            self.account_jws_header = {
                "alg": self.account_key_data["alg"],
                "jwk": self.account_jwk,
            }
            if self.account_uri:
                # Make sure self.account_jws_header is updated
                self.set_account_uri(self.account_uri)

        self.directory = ACMEDirectory(module=module, client=self)

    def set_account_uri(self, uri: str) -> None:
        """
        Set account URI. For ACME v2, it needs to be used to sending signed
        requests.
        """
        self.account_uri = uri
        if self.account_jws_header:
            self.account_jws_header.pop("jwk", None)
            self.account_jws_header["kid"] = self.account_uri

    def parse_key(
        self,
        *,
        key_file: str | os.PathLike | None = None,
        key_content: str | None = None,
        passphrase: str | None = None,
    ) -> dict[str, t.Any]:
        """
        Parses an RSA or Elliptic Curve key file in PEM format and returns key_data.
        In case of an error, raises KeyParsingError.
        """
        if key_file is None and key_content is None:
            raise AssertionError(
                "One of key_file and key_content must be specified!"
            )  # pragma: no cover
        return self.backend.parse_key(
            key_file=key_file, key_content=key_content, passphrase=passphrase
        )

    @t.overload
    def sign_request(
        self,
        *,
        protected: dict[str, t.Any],
        payload: dict[str, t.Any] | None,
        key_data: dict[str, t.Any],
        encode_payload: t.Literal[True] = True,
    ) -> dict[str, t.Any]: ...

    @t.overload
    def sign_request(
        self,
        *,
        protected: dict[str, t.Any],
        payload: str | bytes | None,
        key_data: dict[str, t.Any],
        encode_payload: t.Literal[False],
    ) -> dict[str, t.Any]: ...

    @t.overload
    def sign_request(
        self,
        *,
        protected: dict[str, t.Any],
        payload: str | bytes | dict[str, t.Any] | None,
        key_data: dict[str, t.Any],
        encode_payload: bool = True,
    ) -> dict[str, t.Any]: ...

    def sign_request(
        self,
        *,
        protected: dict[str, t.Any],
        payload: str | bytes | dict[str, t.Any] | None,
        key_data: dict[str, t.Any],
        encode_payload: bool = True,
    ) -> dict[str, t.Any]:
        """
        Signs an ACME request.
        """
        try:
            if payload is None:
                # POST-as-GET
                payload64 = ""
            else:
                # POST
                if encode_payload:
                    payload = self.module.jsonify(payload).encode("utf8")
                payload64 = nopad_b64(to_bytes(payload))
            protected64 = nopad_b64(self.module.jsonify(protected).encode("utf8"))
        except Exception as e:
            raise ModuleFailException(
                f"Failed to encode payload / headers as JSON: {e}"
            ) from e

        return self.backend.sign(
            payload64=payload64, protected64=protected64, key_data=key_data
        )

    def _log(self, msg: str, *, data: t.Any = None) -> None:
        """
        Write arguments to acme.log when logging is enabled.
        """
        if self._debug:
            with open("acme.log", "ab") as f:
                timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%s")
                f.write(f"[{timestamp}] {msg}\n".encode("utf-8"))
                if data is not None:
                    f.write(
                        f"{json.dumps(data, indent=2, sort_keys=True)}\n\n".encode(
                            "utf-8"
                        )
                    )

    @t.overload
    def send_signed_request(
        self,
        url: str,
        payload: dict[str, t.Any] | None,
        *,
        key_data: dict[str, t.Any] | None = None,
        jws_header: dict[str, t.Any] | None = None,
        parse_json_result: t.Literal[True] = True,
        encode_payload: t.Literal[True] = True,
        fail_on_error: bool = True,
        error_msg: str | None = None,
        expected_status_codes: t.Iterable[int] | None = None,
    ) -> tuple[object | bytes, dict[str, t.Any]]: ...

    @t.overload
    def send_signed_request(
        self,
        url: str,
        payload: str | bytes | None,
        *,
        key_data: dict[str, t.Any] | None = None,
        jws_header: dict[str, t.Any] | None = None,
        parse_json_result: t.Literal[True] = True,
        encode_payload: t.Literal[False],
        fail_on_error: bool = True,
        error_msg: str | None = None,
        expected_status_codes: t.Iterable[int] | None = None,
    ) -> tuple[object | bytes, dict[str, t.Any]]: ...

    @t.overload
    def send_signed_request(
        self,
        url: str,
        payload: dict[str, t.Any] | None,
        *,
        key_data: dict[str, t.Any] | None = None,
        jws_header: dict[str, t.Any] | None = None,
        parse_json_result: t.Literal[False],
        encode_payload: t.Literal[True] = True,
        fail_on_error: bool = True,
        error_msg: str | None = None,
        expected_status_codes: t.Iterable[int] | None = None,
    ) -> tuple[bytes, dict[str, t.Any]]: ...

    @t.overload
    def send_signed_request(
        self,
        url: str,
        payload: str | bytes | None,
        *,
        key_data: dict[str, t.Any] | None = None,
        jws_header: dict[str, t.Any] | None = None,
        parse_json_result: t.Literal[False],
        encode_payload: t.Literal[False],
        fail_on_error: bool = True,
        error_msg: str | None = None,
        expected_status_codes: t.Iterable[int] | None = None,
    ) -> tuple[bytes, dict[str, t.Any]]: ...

    def send_signed_request(
        self,
        url: str,
        payload: str | bytes | dict[str, t.Any] | None,
        *,
        key_data: dict[str, t.Any] | None = None,
        jws_header: dict[str, t.Any] | None = None,
        parse_json_result: bool = True,
        encode_payload: bool = True,
        fail_on_error: bool = True,
        error_msg: str | None = None,
        expected_status_codes: t.Iterable[int] | None = None,
    ) -> tuple[object | bytes, dict[str, t.Any]]:
        """
        Sends a JWS signed HTTP POST request to the ACME server and returns
        the response as dictionary (if parse_json_result is True) or in raw form
        (if parse_json_result is False).
        https://tools.ietf.org/html/rfc8555#section-6.2

        If payload is None, a POST-as-GET is performed.
        (https://tools.ietf.org/html/rfc8555#section-6.3)
        """
        key_data = key_data or self.account_key_data
        if key_data is None:
            raise ModuleFailException("Missing key data")
        jws_header = jws_header or self.account_jws_header
        if jws_header is None:
            raise ModuleFailException("Missing JWS header")
        failed_tries = 0
        while True:
            protected = copy.deepcopy(jws_header)
            protected["nonce"] = self.directory.get_nonce()
            protected["url"] = url

            self._log("URL", data=url)
            self._log("protected", data=protected)
            self._log("payload", data=payload)
            data = self.sign_request(
                protected=protected,
                payload=payload,
                key_data=key_data,
                encode_payload=encode_payload,
            )
            self._log("signed request", data=data)
            data_str = self.module.jsonify(data)

            headers = {
                "Content-Type": "application/jose+json",
            }
            resp, info = fetch_url(
                self.module,
                url,
                data=data_str,
                headers=headers,
                method="POST",
                timeout=self.request_timeout,
            )
            if _decode_retry(
                module=self.module, response=resp, info=info, retry_count=failed_tries
            ):
                failed_tries += 1
                continue
            _assert_fetch_url_success(module=self.module, response=resp, info=info)
            result: object | bytes = {}

            try:
                # In Python 2, reading from a closed response yields a TypeError.
                # In Python 3, read() simply returns ''
                if resp.closed:
                    raise TypeError
                content = resp.read()
            except (AttributeError, TypeError):
                content = info.pop("body", None)

            if content or not parse_json_result:
                if (
                    parse_json_result
                    and info["content-type"].startswith("application/json")
                ) or 400 <= info["status"] < 600:
                    try:
                        decoded_result = self.module.from_json(content.decode("utf8"))
                        self._log("parsed result", data=decoded_result)
                        # In case of badNonce error, try again (up to 5 times)
                        # (https://tools.ietf.org/html/rfc8555#section-6.7)
                        if (
                            400 <= info["status"] < 600
                            and failed_tries <= 5
                            and isinstance(decoded_result, dict)
                            and decoded_result.get("type")
                            == "urn:ietf:params:acme:error:badNonce"
                        ):
                            failed_tries += 1
                            continue
                        if parse_json_result:
                            result = decoded_result
                        else:
                            result = content
                    except ValueError as exc:
                        raise NetworkException(
                            f"Failed to parse the ACME response: {url} {content}"
                        ) from exc
                else:
                    result = content

            if fail_on_error and _is_failed(
                info=info, expected_status_codes=expected_status_codes
            ):
                raise ACMEProtocolException(
                    module=self.module,
                    msg=error_msg,
                    info=info,
                    content=content,
                    content_json=result if parse_json_result else None,
                )
            return result, info

    @t.overload
    def get_request(
        self,
        uri: str,
        *,
        parse_json_result: t.Literal[True] = True,
        headers: dict[str, str] | None = None,
        get_only: bool = False,
        fail_on_error: bool = True,
        error_msg: str | None = None,
        expected_status_codes: t.Iterable[int] | None = None,
    ) -> tuple[object, dict[str, t.Any]]: ...

    @t.overload
    def get_request(
        self,
        uri: str,
        *,
        parse_json_result: t.Literal[False],
        headers: dict[str, str] | None = None,
        get_only: bool = False,
        fail_on_error: bool = True,
        error_msg: str | None = None,
        expected_status_codes: t.Iterable[int] | None = None,
    ) -> tuple[bytes, dict[str, t.Any]]: ...

    def get_request(
        self,
        uri: str,
        *,
        parse_json_result: bool = True,
        headers: dict[str, str] | None = None,
        get_only: bool = False,
        fail_on_error: bool = True,
        error_msg: str | None = None,
        expected_status_codes: t.Iterable[int] | None = None,
    ) -> tuple[object | bytes, dict[str, t.Any]]:
        """
        Perform a GET-like request. Will try POST-as-GET for ACMEv2, with fallback
        to GET if server replies with a status code of 405.
        """
        if not get_only:
            # Try POST-as-GET
            content, info = self.send_signed_request(
                uri, None, parse_json_result=False, fail_on_error=False
            )
            if info["status"] == 405:
                # Instead, do unauthenticated GET
                get_only = True
        else:
            # Do unauthenticated GET
            get_only = True

        if get_only:
            # Perform unauthenticated GET
            retry_count = 0
            while True:
                resp, info = fetch_url(
                    self.module,
                    uri,
                    method="GET",
                    headers=headers,
                    timeout=self.request_timeout,
                )
                if not _decode_retry(
                    module=self.module,
                    response=resp,
                    info=info,
                    retry_count=retry_count,
                ):
                    break
                retry_count += 1

            _assert_fetch_url_success(module=self.module, response=resp, info=info)

            try:
                # In Python 2, reading from a closed response yields a TypeError.
                # In Python 3, read() simply returns ''
                if resp.closed:
                    raise TypeError
                content = resp.read()
            except (AttributeError, TypeError):
                content = info.pop("body", None)

        # Process result
        parsed_json_result = False
        result: object | bytes
        if parse_json_result:
            result = {}
            if content:
                if info["content-type"].startswith("application/json"):
                    try:
                        result = self.module.from_json(content.decode("utf8"))
                        parsed_json_result = True
                    except ValueError as exc:
                        raise NetworkException(
                            f"Failed to parse the ACME response: {uri} {content!r}"
                        ) from exc
                else:
                    result = content
        else:
            result = content

        if fail_on_error and _is_failed(
            info=info, expected_status_codes=expected_status_codes
        ):
            raise ACMEProtocolException(
                module=self.module,
                msg=error_msg,
                info=info,
                content=content,
                content_json=(
                    t.cast(dict[str, t.Any], result) if parsed_json_result else None
                ),
            )
        return result, info

    def get_renewal_info(
        self,
        *,
        cert_id: str | None = None,
        cert_info: CertificateInformation | None = None,
        cert_filename: str | os.PathLike | None = None,
        cert_content: str | bytes | None = None,
        include_retry_after: bool = False,
        retry_after_relative_with_timezone: bool = True,
    ) -> dict[str, t.Any]:
        if not self.directory.has_renewal_info_endpoint():
            raise ModuleFailException(
                "The ACME endpoint does not support ACME Renewal Information retrieval"
            )

        if cert_id is None:
            cert_id = compute_cert_id(
                backend=self.backend,
                cert_info=cert_info,
                cert_filename=cert_filename,
                cert_content=cert_content,
            )
        url = f"{self.directory.directory['renewalInfo'].rstrip('/')}/{cert_id}"

        data, info = self.get_request(
            url, parse_json_result=True, fail_on_error=True, get_only=True
        )
        if not isinstance(data, dict):
            raise ACMEProtocolException(
                module=self.module,
                msg="Unexpected renewal information",
                info=info,
                content_json=data,
            )

        # Include Retry-After header if asked for
        if include_retry_after and "retry-after" in info:
            try:
                data["retryAfter"] = parse_retry_after(
                    info["retry-after"],
                    relative_with_timezone=retry_after_relative_with_timezone,
                )
            except ValueError:
                pass
        return data


def create_default_argspec(
    *,
    with_account: bool = True,
    require_account_key: bool = True,
    with_certificate: bool = False,
) -> ArgumentSpec:
    """
    Provides default argument spec for the options documented in the acme doc fragment.
    """
    result = ArgumentSpec(
        argument_spec={
            "acme_directory": {"type": "str", "required": True},
            "acme_version": {"type": "int", "choices": [2], "default": 2},
            "validate_certs": {"type": "bool", "default": True},
            "select_crypto_backend": {
                "type": "str",
                "default": "auto",
                "choices": ["auto", "openssl", "cryptography"],
            },
            "request_timeout": {"type": "int", "default": 10},
        },
    )
    if with_account:
        result.update_argspec(
            account_key_src={"type": "path", "aliases": ["account_key"]},
            account_key_content={"type": "str", "no_log": True},
            account_key_passphrase={"type": "str", "no_log": True},
            account_uri={"type": "str"},
        )
        if require_account_key:
            result.update(required_one_of=[["account_key_src", "account_key_content"]])
        result.update(mutually_exclusive=[["account_key_src", "account_key_content"]])
    if with_certificate:
        result.update_argspec(
            csr={"type": "path"},
            csr_content={"type": "str"},
        )
        result.update(
            required_one_of=[["csr", "csr_content"]],
            mutually_exclusive=[["csr", "csr_content"]],
        )
    return result


def create_backend(
    module: AnsibleModule, *, needs_acme_v2: bool = True
) -> CryptoBackend:
    backend = module.params["select_crypto_backend"]

    # Backend autodetect
    if backend == "auto":
        backend = "cryptography" if HAS_CURRENT_CRYPTOGRAPHY else "openssl"

    # Create backend object
    module_backend: CryptoBackend
    if backend == "cryptography":
        if CRYPTOGRAPHY_ERROR is not None:
            # Either we could not import cryptography at all, or there was an unexpected error
            if CRYPTOGRAPHY_VERSION is None:
                msg = missing_required_lib("cryptography")
            else:
                msg = f"Unexpected error while preparing cryptography: {CRYPTOGRAPHY_ERROR.splitlines()[-1]}"
            module.fail_json(msg=msg, exception=CRYPTOGRAPHY_ERROR)
        if not HAS_CURRENT_CRYPTOGRAPHY:
            # We succeeded importing cryptography, but its version is too old.
            mrl = missing_required_lib(
                f"cryptography >= {CRYPTOGRAPHY_MINIMAL_VERSION}"
            )
            module.fail_json(
                msg=f"Found cryptography, but only version {CRYPTOGRAPHY_VERSION}. {mrl}"
            )
        module.debug(
            f"Using cryptography backend (library version {CRYPTOGRAPHY_VERSION})"
        )
        module_backend = CryptographyBackend(module=module)
    elif backend == "openssl":
        module.debug("Using OpenSSL binary backend")
        module_backend = OpenSSLCLIBackend(module=module)
    else:
        module.fail_json(msg=f'Unknown crypto backend "{backend}"!')

    # Check common module parameters
    if not module.params["validate_certs"]:
        module.warn(
            "Disabling certificate validation for communications with ACME endpoint. "
            "This should only be done for testing against a local ACME server for "
            "development purposes, but *never* for production purposes."
        )

    # AnsibleModule() changes the locale, so change it back to C because we rely
    # on datetime.datetime.strptime() when parsing certificate dates.
    locale.setlocale(locale.LC_ALL, "C")

    return module_backend


__all__ = (
    "ACMEClient",
    "ACMEDirectory",
    "create_backend",
    "create_default_argspec",
)



# Copyright (c) 2017-2020 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t

from ansible.module_utils.common.text.converters import to_native

try:
    import lxml.etree

    HAS_LXML_ETREE = True
except ImportError:
    HAS_LXML_ETREE = False

from ansible_collections.community.dns.plugins.module_utils._http import NetworkError

if t.TYPE_CHECKING:
    from collections.abc import Mapping  # pragma: no cover

    from ._http import HTTPHelper  # pragma: no cover


class WSDLException(Exception):
    pass


class WSDLNetworkError(WSDLException):
    pass


class WSDLError(WSDLException):
    def __init__(self, origin, error_code, message):
        super().__init__(f"{origin} ({error_code}): {message}")
        self.error_origin = origin
        self.error_code = error_code
        self.error_message = message


class WSDLCodingException(WSDLException):
    pass


def _split_text_namespace(
    node: lxml.etree.Element, text: str
) -> tuple[str, str | None]:
    i = text.find(":")
    if i < 0:
        return text, None
    ns = node.nsmap.get(text[:i])
    text = text[i + 1 :]
    return text, ns


_NAMESPACE_ENVELOPE = "http://schemas.xmlsoap.org/soap/envelope/"
_NAMESPACE_XSI = "http://www.w3.org/2001/XMLSchema-instance"
_NAMESPACE_XSD = "http://www.w3.org/2001/XMLSchema"
_NAMESPACE_XML_SOAP = "http://xml.apache.org/xml-soap"
_NAMESPACE_XML_SOAP_ENCODING = "http://schemas.xmlsoap.org/soap/encoding/"


def _set_type(
    node: lxml.etree.Element, type_value: str, namespace: str | None = None
) -> None:
    tv: lxml.etree.QName | str
    if namespace is not None:
        tv = lxml.etree.QName(namespace, type_value)
    else:
        tv = type_value
    node.set(lxml.etree.QName(_NAMESPACE_XSI, "type").text, tv)


def encode_wsdl(node: lxml.etree.Element, value: t.Any) -> None:
    if value is None:
        node.set(lxml.etree.QName(_NAMESPACE_XSI, "nil").text, "true")
    elif isinstance(value, str):
        _set_type(node, "xsd:string")
        node.text = value
    elif isinstance(value, int):
        _set_type(node, "xsd:int")
        node.text = str(value)
    elif isinstance(value, bool):
        _set_type(node, "xsd:boolean")
        node.text = "true" if value else "false"
    elif isinstance(value, dict):
        _set_type(node, "Map", _NAMESPACE_XML_SOAP)
        for key, val in sorted(value.items()):
            child = lxml.etree.Element("item")
            ke = lxml.etree.Element("key")
            encode_wsdl(ke, key)
            child.append(ke)
            ve = lxml.etree.Element("value")
            encode_wsdl(ve, val)
            child.append(ve)
            node.append(child)
    elif isinstance(value, list):
        _set_type(node, "SOAP-ENC:Array")
        for elt in value:
            child = lxml.etree.Element("item")
            encode_wsdl(child, elt)
            node.append(child)
    else:
        raise WSDLCodingException(f"Do not know how to encode {type(value)}!")


def _decode_wsdl_array(
    result: list[t.Any],
    node: lxml.etree.Element,
    root_ns: str | None,
    ids: dict[str, str],
):
    for item in node:
        if item.tag != "item":
            raise WSDLCodingException(f"Invalid child tag {item.tag!r} in map!")
        result.append(decode_wsdl(item, root_ns, ids))


def decode_wsdl(
    node: lxml.etree.Element, root_ns: str | None, ids: dict[str, str]
) -> t.Any:
    href = node.get("href")
    nil = node.get(lxml.etree.QName(_NAMESPACE_XSI, "nil"))
    nid = node.get("id")
    result: t.Any
    if href is not None:
        if not href.startswith("#"):
            raise WSDLCodingException(f'Global reference "{href}" not supported!')
        href = href[1:]
        if href not in ids:
            raise WSDLCodingException(f'ID "{href}" not yet defined!')
        result = ids[href]
    elif nil == "true":
        result = None
    else:
        type_with_ns = node.get(lxml.etree.QName(_NAMESPACE_XSI, "type"))
        if type_with_ns is None:
            raise WSDLCodingException(f'Element "{node}" has no "xsi:type" tag!')
        ntype, ns = _split_text_namespace(node, type_with_ns)
        if ns is None:
            raise WSDLCodingException(f'Cannot find namespace for "{type_with_ns}"!')
        if ns == _NAMESPACE_XSD:
            if ntype == "boolean":
                if node.text == "true":
                    result = True
                elif node.text == "false":
                    result = False
                else:
                    raise WSDLCodingException(
                        f"Invalid value for boolean: {node.text!r}"
                    )
            elif ntype == "int":
                try:
                    if node.text is None:
                        raise TypeError
                    result = int(node.text)
                except (TypeError, ValueError) as exc:
                    raise WSDLCodingException(
                        f"Invalid value for int: {node.text!r}"
                    ) from exc
            elif ntype == "string":
                result = node.text
            else:
                raise WSDLCodingException(f'Unknown XSD type "{ntype}"!')
        elif ns == _NAMESPACE_XML_SOAP:
            if ntype == "Map":
                result = {}
                if nid is not None:
                    ids[nid] = result
                for item in node:
                    if item.tag != "item":
                        raise WSDLCodingException(
                            f"Invalid child tag {item.tag!r} in map!"
                        )
                    key = item.find("key")
                    if key is None:
                        raise WSDLCodingException(f"Cannot find key for {item!r}!")
                    key = decode_wsdl(key, root_ns, ids)
                    value = item.find("value")
                    if value is None:
                        raise WSDLCodingException(f"Cannot find value for {item!r}!")
                    value = decode_wsdl(value, root_ns, ids)
                    result[key] = value
                return result
            raise WSDLCodingException(f'Unknown XSD type "{ntype}"!')
        elif ns == _NAMESPACE_XML_SOAP_ENCODING:
            if ntype == "Array":
                result = []
                if nid is not None:
                    ids[nid] = result
                _decode_wsdl_array(result, node, root_ns, ids)
            else:
                raise WSDLCodingException(f'Unknown XSD type "{ntype}"!')
        elif ns == root_ns:
            array_type = node.get(
                lxml.etree.QName(_NAMESPACE_XML_SOAP_ENCODING, "arrayType")
            )
            if array_type is not None:
                result = []
                if nid is not None:
                    ids[nid] = result
                _decode_wsdl_array(result, node, root_ns, ids)
            else:
                result = {}
                if nid is not None:
                    ids[nid] = result
                for item in node:
                    result[item.tag] = decode_wsdl(item, root_ns, ids)
        else:
            raise WSDLCodingException(
                f'Unknown type namespace "{ns}" (with type "{ntype}")!'
            )
    if nid is not None:
        ids[nid] = result
    return result


class Parser:
    def _parse(
        self, result: dict[str, t.Any], node: lxml.etree.Element, where: str
    ) -> None:
        for child in node:
            assert isinstance(child.tag, (str, bytes, lxml.etree.QName))
            tag = lxml.etree.QName(child.tag)
            if tag.namespace != self._api:
                raise WSDLCodingException(
                    f'Cannot interpret {where} item of type "{tag}"!'
                )
            for res in child.iter("return"):
                result[tag.localname] = decode_wsdl(res, self._api, {})

    def __init__(self, api: str, root: lxml.etree.Element) -> None:
        self._main_ns = _NAMESPACE_ENVELOPE
        self._api = api
        self._root = root
        for fault in self._root.iter(lxml.etree.QName(self._main_ns, "Fault").text):
            fault_code = fault.find("faultcode")
            fault_code_val = None
            fault_string = fault.find("faultstring")
            origin = "server"
            if fault_code is not None and fault_code.text:
                code, code_ns = _split_text_namespace(fault, fault_code.text)
                fault_code_val = code
                if code_ns == self._main_ns:
                    origin = code.lower()
            if fault_string is not None and fault_string.text:
                raise WSDLError(origin, fault_code_val, fault_string.text)
            raise WSDLError(
                origin, fault_code_val, lxml.etree.tostring(fault).decode("utf-8")
            )
        self._header: dict[str, t.Any] = {}
        self._body: dict[str, t.Any] = {}
        for header in self._root.iter(lxml.etree.QName(self._main_ns, "Header").text):
            self._parse(self._header, header, "header")
        for body in self._root.iter(lxml.etree.QName(self._main_ns, "Body").text):
            self._parse(self._body, body, "body")

    def get_header(self, header: str) -> t.Any:
        return self._header[header]

    def get_result(self, body: str) -> t.Any:
        return self._body[body]

    def __str__(self) -> str:
        return f"header={self._header}, body={self._body}"

    def __repr__(self) -> str:
        return (
            """<?xml version='1.0' encoding='utf-8'?>"""
            + "\n"
            + lxml.etree.tostring(self._root, pretty_print=True).decode("utf-8")
        )


class Composer:
    @staticmethod
    def _create(tag: str, namespace: str | None = None, **kwarg) -> lxml.etree.Element:
        if namespace:
            return lxml.etree.Element(lxml.etree.QName(namespace, tag), **kwarg)
        return lxml.etree.Element(tag, **kwarg)

    def __str__(self) -> str:
        return (
            """<?xml version='1.0' encoding='utf-8'?>"""
            + "\n"
            + lxml.etree.tostring(self._root, pretty_print=True).decode("utf-8")
        )

    def _create_envelope(self, tag: str, **kwarg) -> lxml.etree.Element:
        return self._create(tag, self._main_ns, **kwarg)

    def __init__(
        self,
        http_helper: HTTPHelper,
        api: str,
        namespaces: Mapping[str, str] | None = None,
    ) -> None:
        self._http_helper = http_helper
        self._main_ns = _NAMESPACE_ENVELOPE
        self._api = api
        # Compose basic document
        all_namespaces = {
            "SOAP-ENV": _NAMESPACE_ENVELOPE,
            "xsd": _NAMESPACE_XSD,
            "xsi": _NAMESPACE_XSI,
            "ns2": "auth",
            "SOAP-ENC": _NAMESPACE_XML_SOAP_ENCODING,
        }
        if namespaces is not None:
            all_namespaces.update(namespaces)
        self._root = self._create_envelope("Envelope", nsmap=all_namespaces)
        self._root.set(
            lxml.etree.QName(self._main_ns, "encodingStyle").text,
            _NAMESPACE_XML_SOAP_ENCODING,
        )
        self._header = self._create_envelope("Header")
        self._root.append(self._header)
        self._body = self._create_envelope("Body")
        self._root.append(self._body)
        self._command: str | None = None

    def add_auth(self, username: str, password: str) -> None:
        auth = self._create("authenticate", "auth")
        user = self._create("UserName")
        user.text = username
        auth.append(user)
        pw = self._create("Password")
        pw.text = password
        auth.append(pw)
        self._header.append(auth)

    def add_simple_command(self, command: str, **args) -> None:
        self._command = command
        cmd = self._create(command, self._api)
        for arg, value in args.items():
            arg_elt = self._create(arg)
            encode_wsdl(arg_elt, value)
            cmd.append(arg_elt)
        self._body.append(cmd)

    def execute(self, debug: bool = False) -> Parser:
        payload = (
            b"""<?xml version='1.0' encoding='utf-8'?>"""
            + b"\n"
            + lxml.etree.tostring(self._root)
            + b"\n"
        )
        try:
            headers = {
                "Content-Type": "text/xml; charset=utf-8",
                "Content-Length": str(len(payload)),
            }
            if self._command:
                headers["SOAPAction"] = f'"{self._api}#{self._command}"'
            result, info = self._http_helper.fetch_url(
                self._api, data=payload, method="POST", timeout=300, headers=headers
            )
            code = info["status"]
        except NetworkError as e:
            raise WSDLNetworkError(to_native(e)) from e
        # if debug:
        #     q.q('Result: {0}, content: {1}'.format(code, result.decode('utf-8')))
        try:
            node = lxml.etree.fromstring(result or b"")
        except Exception as exc:
            raise WSDLError(
                "server",
                "",
                f"Error while parsing WSDL command result: {exc}\nResult:\n{(result or b'').decode('utf-8')}",
            ) from exc
        parser = Parser(self._api, node)
        if code < 200 or code >= 300:
            raise WSDLError(
                "server",
                "",
                f"Error {code} while executing WSDL command:\n{(result or b'').decode('utf-8')}",
            )
        return parser



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_switch_stacks_routing_static_routes
short_description: Resource module for networks _switch _stacks _routing _static _routes
description:
  - Manage operations create, update and delete of the resource networks _switch _stacks
    _routing _static _routes.
  - Create a layer 3 static route for a switch stack.
  - Delete a layer 3 static route for a switch stack.
  - Update a layer 3 static route for a switch stack.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  advertiseViaOspfEnabled:
    description: Option to advertise static route via OSPF.
    type: bool
  managementNextHop:
    description: Optional fallback IP address for management traffic.
    type: str
  name:
    description: Name or description for layer 3 static route.
    type: str
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
  nextHopIp:
    description: IP address of the next hop device to which the device sends its traffic
      for the subnet.
    type: str
  preferOverOspfRoutesEnabled:
    description: Option to prefer static route over OSPF routes.
    type: bool
  staticRouteId:
    description: StaticRouteId path parameter. Static route ID.
    type: str
  subnet:
    description: The subnet which is routed via this static route and should be specified
      in CIDR notation (ex. 1.2.3.0/24).
    type: str
  switchStackId:
    description: SwitchStackId path parameter. Switch stack ID.
    type: str
  vrf:
    description: The VRF settings of the interface. Requires IOS XE 17.18 or higher.
    suboptions:
      leakRouteToDefaultVrf:
        description: Whether or not next-hop IP is reachable via default VRF.
        type: bool
      name:
        description: The name of the VRF this static route belongs to.
        type: str
    type: dict
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for switch createNetworkSwitchStackRoutingStaticRoute
    description: Complete reference of the createNetworkSwitchStackRoutingStaticRoute
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-network-switch-stack-routing-static-route
  - name: Cisco Meraki documentation for switch deleteNetworkSwitchStackRoutingStaticRoute
    description: Complete reference of the deleteNetworkSwitchStackRoutingStaticRoute
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!delete-network-switch-stack-routing-static-route
  - name: Cisco Meraki documentation for switch updateNetworkSwitchStackRoutingStaticRoute
    description: Complete reference of the updateNetworkSwitchStackRoutingStaticRoute
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-network-switch-stack-routing-static-route
notes:
  - SDK Method used are
    switch.Switch.create_network_switch_stack_routing_static_route,
    switch.Switch.delete_network_switch_stack_routing_static_route,
    switch.Switch.update_network_switch_stack_routing_static_route,
  - Paths used are
    post /networks/{networkId}/switch/stacks/{switchStackId}/routing/staticRoutes,
    delete /networks/{networkId}/switch/stacks/{switchStackId}/routing/staticRoutes/{staticRouteId},
    put /networks/{networkId}/switch/stacks/{switchStackId}/routing/staticRoutes/{staticRouteId},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.networks_switch_stacks_routing_static_routes:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    advertiseViaOspfEnabled: false
    name: My route
    networkId: string
    nextHopIp: 1.2.3.4
    preferOverOspfRoutesEnabled: false
    subnet: 192.168.1.0/24
    switchStackId: string
    vrf:
      leakRouteToDefaultVrf: false
      name: Blue
- name: Update by id
  cisco.meraki.networks_switch_stacks_routing_static_routes:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    advertiseViaOspfEnabled: false
    managementNextHop: 1.2.3.5
    name: My route
    networkId: string
    nextHopIp: 1.2.3.4
    preferOverOspfRoutesEnabled: false
    staticRouteId: string
    subnet: 192.168.1.0/24
    switchStackId: string
    vrf:
      leakRouteToDefaultVrf: false
      name: Blue
- name: Delete by id
  cisco.meraki.networks_switch_stacks_routing_static_routes:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: absent
    networkId: string
    staticRouteId: string
    switchStackId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "staticRouteId": "string",
      "name": "string",
      "subnet": "string",
      "nextHopIp": "string",
      "managementNextHop": {},
      "vrf": {
        "name": "string",
        "leakRouteToDefaultVrf": true
      },
      "advertiseViaOspfEnabled": true,
      "preferOverOspfRoutesEnabled": true
    }
"""



#!/usr/bin/python
# Copyright (c) 2026 Shreyash Bhosale <shreyashpb16@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: golang_package
short_description: Manage Go packages with C(go install)
version_added: 13.1.0
description:
  - Manage Go packages with C(go install).
  - Packages are installed as binaries into the Go binary directory (E(GOBIN)).
  - Uninstalling a package removes its binary from E(GOBIN).
author: "Shreyash Bhosale (@shrbhosa)"
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  executable:
    description:
      - Path to the C(go) binary.
      - If not specified, the module looks for C(go) in E(PATH).
    type: path
  name:
    description:
      - The name of a Go package to install.
      - Must be a full Go package path, for example V(golang.org/x/tools/cmd/stringer).
      - A version can be specified inline using the V(@) suffix, for example
        V(golang.org/x/tools/cmd/stringer@v0.29.0). This allows pinning different
        versions when installing multiple packages.
      - The inline V(@version) syntax cannot be combined with the O(version) parameter.
    type: list
    elements: str
    required: true
  version:
    description:
      - The version to install, for example V(v1.55.2).
      - The version must include the V(v) prefix as required by Go module versioning.
      - Can only be used when O(name) contains a single package without an inline V(@version).
      - When not specified and O(state=present), the latest version is installed for new packages.
    type: str
  state:
    description:
      - The desired state of the Go package.
    type: str
    default: present
    choices: ["present", "absent", "latest"]
requirements:
  - Go >= 1.16
notes:
  - The Go binary directory is determined by C(go env GOBIN). To customize the install location,
    set the E(GOBIN) environment variable using the task's C(environment) directive.
  - Go does not provide a native uninstall command. When O(state=absent), the module removes the
    binary file from E(GOBIN) directly.
"""

EXAMPLES = r"""
- name: Install stringer
  community.general.golang_package:
    name: golang.org/x/tools/cmd/stringer

- name: Install golangci-lint at a specific version
  community.general.golang_package:
    name: github.com/golangci/golangci-lint/cmd/golangci-lint
    version: v1.55.2

- name: Install multiple packages at specific versions
  community.general.golang_package:
    name:
      - golang.org/x/tools/cmd/stringer@v0.29.0
      - github.com/golangci/golangci-lint/cmd/golangci-lint@v1.55.2

- name: Install multiple packages
  community.general.golang_package:
    name:
      - golang.org/x/tools/cmd/stringer
      - golang.org/x/tools/cmd/goimports

- name: Remove stringer
  community.general.golang_package:
    name: golang.org/x/tools/cmd/stringer
    state: absent

- name: Update golangci-lint to the latest version
  community.general.golang_package:
    name: github.com/golangci/golangci-lint/cmd/golangci-lint
    state: latest

- name: Install into a custom directory
  community.general.golang_package:
    name: golang.org/x/tools/cmd/stringer
  environment:
    GOBIN: /usr/local/bin
"""

import json
import os

from ansible.module_utils.basic import AnsibleModule


class Go:
    def __init__(self, module, **kwargs):
        self.module = module
        self.executable = [kwargs["executable"] or module.get_bin_path("go", True)]
        self.state = kwargs["state"]
        self.version = kwargs["version"]
        self._gobin = None

        self.packages = []
        for name in kwargs["name"]:
            if "@" in name:
                pkg, ver = name.rsplit("@", 1)
                self.packages.append((pkg, ver))
            else:
                self.packages.append((name, None))

    @property
    def package_names(self):
        return [pkg for pkg, dummy in self.packages]

    def _exec(self, args, run_in_check_mode=False, check_rc=True):
        if not self.module.check_mode or run_in_check_mode:
            cmd = self.executable + args
            rc, out, err = self.module.run_command(cmd, check_rc=check_rc)
            return out, err
        return "", ""

    def _get_gobin(self):
        if self._gobin is not None:
            return self._gobin

        out, dummy = self._exec(["env", "GOBIN"], run_in_check_mode=True, check_rc=True)
        gobin = out.strip()
        if gobin:
            self._gobin = gobin
            return self._gobin

        out, dummy = self._exec(["env", "GOPATH"], run_in_check_mode=True, check_rc=True)
        gopath = out.strip()
        if not gopath:
            self.module.fail_json(msg="Could not determine Go binary directory: GOBIN and GOPATH are both empty")
        self._gobin = os.path.join(gopath.split(os.pathsep)[0], "bin")
        return self._gobin

    def _get_binary_info(self, binary_path):
        out, dummy = self._exec(["version", "-m", binary_path], run_in_check_mode=True, check_rc=False)
        if not out:
            return None

        path = None
        version = None
        mod_path = None
        for line in out.splitlines():
            parts = line.strip().split("\t")
            if len(parts) >= 2 and parts[0] == "path":
                path = parts[1]
            elif len(parts) >= 3 and parts[0] == "mod":
                mod_path = parts[1]
                version = parts[2]

        if path and version:
            return {"path": path, "version": version, "mod_path": mod_path}
        return None

    def get_installed(self):
        gobin = self._get_gobin()
        installed = {}
        for pkg_name in self.package_names:
            binary_name = pkg_name.rsplit("/", 1)[-1]
            binary_path = os.path.join(gobin, binary_name)
            if not os.path.isfile(binary_path):
                continue
            info = self._get_binary_info(binary_path)
            if info and info["path"] == pkg_name:
                installed[pkg_name] = info["version"]
        return installed

    def _get_latest_version(self, mod_path):
        out, dummy = self._exec(
            ["list", "-m", "-json", mod_path + "@latest"],
            run_in_check_mode=True,
            check_rc=False,
        )
        if out:
            try:
                data = json.loads(out)
                return data.get("Version")
            except (json.JSONDecodeError, ValueError):
                pass
        return None

    def is_outdated(self, package):
        gobin = self._get_gobin()
        binary_name = package.rsplit("/", 1)[-1]
        binary_path = os.path.join(gobin, binary_name)
        info = self._get_binary_info(binary_path)
        if not info or info["path"] != package:
            return True
        installed_version = info["version"]
        mod_path = info.get("mod_path")
        if not mod_path:
            self.module.fail_json(msg=f"Could not determine module path for package {package}")
        latest_version = self._get_latest_version(mod_path)
        if not latest_version:
            self.module.fail_json(msg=f"Could not determine latest version for package {package}")
        return installed_version != latest_version

    def install(self, packages):
        by_version = {}
        for pkg, ver in packages:
            effective = ver or self.version or "latest"
            by_version.setdefault(effective, []).append(pkg)

        out_all, err_all = "", ""
        for ver, pkgs in by_version.items():
            cmd = ["install"] + [p + "@" + ver for p in pkgs]
            out, err = self._exec(cmd)
            out_all += out
            err_all += err
        return out_all, err_all

    def uninstall(self, packages):
        gobin = self._get_gobin()
        for package in packages:
            binary_name = package.rsplit("/", 1)[-1]
            binary_path = os.path.join(gobin, binary_name)
            if os.path.isfile(binary_path) and not self.module.check_mode:
                os.remove(binary_path)


def main():
    arg_spec = dict(
        executable=dict(type="path"),
        name=dict(required=True, type="list", elements="str"),
        state=dict(default="present", choices=["present", "absent", "latest"]),
        version=dict(type="str"),
    )
    module = AnsibleModule(argument_spec=arg_spec, supports_check_mode=True)

    name = module.params["name"]
    state = module.params["state"]
    version = module.params["version"]

    has_inline_version = any("@" in n for n in name)
    if version and has_inline_version:
        module.fail_json(msg="Cannot combine the 'version' parameter with inline @version in package names")
    if version and len(name) > 1:
        module.fail_json(msg="The 'version' parameter can only be used when 'name' contains a single package")
    if state == "latest" and (version or has_inline_version):
        module.fail_json(msg="Cannot specify a version when state=latest")

    module.run_command_environ_update = dict(LANGUAGE="C", LC_ALL="C")

    go = Go(module, **module.params)
    changed, out, err = False, None, None
    installed_packages = go.get_installed()

    if state == "present":
        to_install = []
        for pkg, ver in go.packages:
            effective_ver = ver or version
            if pkg not in installed_packages:
                to_install.append((pkg, ver))
            elif effective_ver and effective_ver != installed_packages[pkg]:
                to_install.append((pkg, ver))
        if to_install:
            changed = True
            out, err = go.install(to_install)
    elif state == "latest":
        to_update = [(pkg, ver) for pkg, ver in go.packages if pkg not in installed_packages or go.is_outdated(pkg)]
        if to_update:
            changed = True
            out, err = go.install(to_update)
    else:  # absent
        to_uninstall = [pkg for pkg in go.package_names if pkg in installed_packages]
        if to_uninstall:
            changed = True
            go.uninstall(to_uninstall)

    module.exit_json(changed=changed, stdout=out, stderr=err)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the
# cli_rm_builder.
#
# Manually editing this file is not advised.
#
# To update the argspec make the desired changes
# in the module docstring and re-run
# cli_rm_builder.
#
#############################################

"""
The arg spec for the nxos_bgp_address_family module
"""


class Bgp_address_familyArgs(object):  # pylint: disable=R0903
    """The arg spec for the nxos_bgp_address_family module"""

    argument_spec = {
        "running_config": {"type": "str"},
        "config": {
            "type": "dict",
            "options": {
                "as_number": {"type": "str"},
                "address_family": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "afi": {
                            "type": "str",
                            "choices": [
                                "ipv4",
                                "ipv6",
                                "link-state",
                                "vpnv4",
                                "vpnv6",
                                "l2vpn",
                            ],
                            "required": True,
                        },
                        "safi": {
                            "type": "str",
                            "choices": [
                                "unicast",
                                "multicast",
                                "mvpn",
                                "evpn",
                            ],
                        },
                        "additional_paths": {
                            "type": "dict",
                            "options": {
                                "install_backup": {"type": "bool"},
                                "receive": {"type": "bool"},
                                "selection": {
                                    "type": "dict",
                                    "options": {"route_map": {"type": "str"}},
                                },
                                "send": {"type": "bool"},
                            },
                        },
                        "advertise_l2vpn_evpn": {"type": "bool"},
                        "advertise_pip": {"type": "bool"},
                        "advertise_system_mac": {"type": "bool"},
                        "allow_vni_in_ethertag": {"type": "bool"},
                        "aggregate_address": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "prefix": {"type": "str"},
                                "advertise_map": {"type": "str"},
                                "as_set": {"type": "bool"},
                                "attribute_map": {"type": "str"},
                                "summary_only": {"type": "bool"},
                                "suppress_map": {"type": "str"},
                            },
                        },
                        "client_to_client": {
                            "type": "dict",
                            "options": {"no_reflection": {"type": "bool"}},
                        },
                        "dampen_igp_metric": {"type": "int"},
                        "dampening": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "decay_half_life": {"type": "int"},
                                "start_reuse_route": {"type": "int"},
                                "start_suppress_route": {"type": "int"},
                                "max_suppress_time": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "default_information": {
                            "type": "dict",
                            "options": {"originate": {"type": "bool"}},
                        },
                        "default_metric": {"type": "int"},
                        "distance": {
                            "type": "dict",
                            "options": {
                                "ebgp_routes": {"type": "int"},
                                "ibgp_routes": {"type": "int"},
                                "local_routes": {"type": "int"},
                            },
                        },
                        "export_gateway_ip": {"type": "bool"},
                        "inject_map": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "route_map": {"type": "str"},
                                "exist_map": {"type": "str"},
                                "copy_attributes": {"type": "bool"},
                            },
                        },
                        "maximum_paths": {
                            "type": "dict",
                            "options": {
                                "parallel_paths": {"type": "int"},
                                "ibgp": {
                                    "type": "dict",
                                    "options": {"parallel_paths": {"type": "int"}},
                                },
                                "eibgp": {
                                    "type": "dict",
                                    "options": {"parallel_paths": {"type": "int"}},
                                },
                                "local": {
                                    "type": "dict",
                                    "options": {"parallel_paths": {"type": "int"}},
                                },
                                "mixed": {
                                    "type": "dict",
                                    "options": {"parallel_paths": {"type": "int"}},
                                },
                            },
                        },
                        "networks": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "prefix": {"type": "str"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "nexthop": {
                            "type": "dict",
                            "options": {
                                "route_map": {"type": "str"},
                                "trigger_delay": {
                                    "type": "dict",
                                    "options": {
                                        "critical_delay": {"type": "int"},
                                        "non_critical_delay": {"type": "int"},
                                    },
                                },
                            },
                        },
                        "redistribute": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "protocol": {
                                    "type": "str",
                                    "choices": [
                                        "am",
                                        "direct",
                                        "eigrp",
                                        "isis",
                                        "lisp",
                                        "ospf",
                                        "ospfv3",
                                        "rip",
                                        "static",
                                        "hmm",
                                    ],
                                    "required": True,
                                },
                                "id": {"type": "str"},
                                "route_map": {"type": "str", "required": True},
                            },
                        },
                        "retain": {
                            "type": "dict",
                            "options": {
                                "route_target": {
                                    "type": "dict",
                                    "options": {
                                        "retain_all": {"type": "bool"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                            },
                        },
                        "suppress_inactive": {"type": "bool"},
                        "table_map": {
                            "type": "dict",
                            "options": {
                                "name": {"type": "str", "required": True},
                                "filter": {"type": "bool"},
                            },
                        },
                        "timers": {
                            "type": "dict",
                            "options": {
                                "bestpath_defer": {
                                    "type": "dict",
                                    "options": {
                                        "defer_time": {"type": "int"},
                                        "maximum_defer_time": {"type": "int"},
                                    },
                                },
                            },
                        },
                        "wait_igp_convergence": {"type": "bool"},
                        "vrf": {"type": "str"},
                    },
                },
            },
        },
        "state": {
            "type": "str",
            "choices": [
                "merged",
                "replaced",
                "overridden",
                "deleted",
                "parsed",
                "gathered",
                "rendered",
            ],
            "default": "merged",
        },
    }  # pylint: disable=C0301



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Dag Wieers (@dagwieers) <dag@wieers.com>
# Copyright: (c) 2021, Anvitha Jain (@anvitha-jain) <anvjain@cisco.com>
# Copyright: (c) 2022, Akini Ross (@akinross) <akinross@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_site_bd_l3out
short_description: Manage site-local BD l3out's in schema template
description:
- Manage site-local BDs l3out's in schema template on Cisco ACI Multi-Site.
author:
- Dag Wieers (@dagwieers)
- Anvitha Jain (@anvitha-jain)
- Akini Ross (@akinross)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  site:
    description:
    - The name of the site.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
  bd:
    description:
    - The name of the BD.
    type: str
    required: true
  l3out:
    description:
    - The l3out associated to this BD.
    aliases: [ name ]
    type: dict
    suboptions:
      name:
        description:
        - The name of the l3out to associate with.
        required: true
        type: str
      schema:
        description:
        - The schema that defines the referenced l3out.
        - If this parameter is unspecified, it defaults to the current schema.
        - Mutually exclusive with O(l3out.tenant).
        type: str
      template:
        description:
        - The template that defines the referenced l3out.
        - If this parameter is unspecified, it defaults to the current schema.
        - Mutually exclusive with O(l3out.tenant).
        type: str
      tenant:
        description:
        - The tenant name of the referenced l3out.
        - If this parameter is specified, the constructed l3out reference will refer to a distinguished name (DN) in APIC.
        - Mutually exclusive with O(l3out.schema) and O(l3out.template).
        type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
notes:
- The ACI MultiSite PATCH API has a deficiency requiring some objects to be referenced by index.
  This can cause silent corruption on concurrent access when changing/removing on object as
  the wrong object may be referenced. This module is affected by this deficiency.
seealso:
- module: cisco.mso.mso_schema_site_bd
- module: cisco.mso.mso_schema_template_bd
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a new site BD l3out
  cisco.mso.mso_schema_site_bd_l3out:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    bd: BD1
    l3out:
      name: L3out1
    state: present

- name: Add a new site BD l3out with different schema and template
  cisco.mso.mso_schema_site_bd_l3out:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    bd: BD1
    l3out:
      name: L3out1
      schema: Schema2
      template: Template2
    state: present

- name: Remove a site BD l3out
  cisco.mso.mso_schema_site_bd_l3out:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    bd: BD1
    l3out:
      name: L3out1
    state: absent

- name: Query a specific site BD l3out
  cisco.mso.mso_schema_site_bd_l3out:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    bd: BD1
    l3out:
      name: L3out1
    state: query
  register: query_result

- name: Query all site BD l3outs
  cisco.mso.mso_schema_site_bd_l3out:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    bd: BD1
    state: query
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec, mso_l3out_reference_spec
from ansible_collections.cisco.mso.plugins.module_utils.schema import MSOSchema


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        site=dict(type="str", required=True),
        template=dict(type="str", required=True),
        bd=dict(type="str", required=True),
        l3out=dict(
            type="dict", options=mso_l3out_reference_spec(), aliases=["name"], mutually_exclusive=[("tenant", "schema"), ("tenant", "template")]
        ),  # This parameter is not required for querying all objects
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["l3out"]],
            ["state", "present", ["l3out"]],
        ],
    )

    schema = module.params.get("schema")
    site = module.params.get("site")
    template = module.params.get("template").replace(" ", "")
    bd = module.params.get("bd")
    l3out = module.params.get("l3out")
    state = module.params.get("state")

    mso = MSOModule(module)

    mso_schema = MSOSchema(mso, schema, template, site)
    mso_objects = mso_schema.schema_objects

    mso_schema.set_template_bd(bd)
    mso_schema.set_site_bd(bd, fail_module=False)

    bd_path = "/sites/{0}-{1}/bds".format(mso_objects.get("site").details.get("siteId"), template)
    ops = []
    payload = dict()

    if l3out:

        if l3out.get("tenant"):
            l3out_ref = "uni/tn-{0}/out-{1}".format(l3out.get("tenant"), l3out.get("name"))
        else:
            l3out_schema_id = mso.lookup_schema(l3out.get("schema")) if l3out.get("schema") else mso_schema.id
            l3out_template = l3out.get("template") if l3out.get("template") else template
            l3out_ref = mso.l3out_ref(schema_id=l3out_schema_id, template=l3out_template, l3out=l3out.get("name"))

        if not mso_objects.get("site_bd"):
            payload = dict(bdRef=dict(schemaId=mso_schema.id, templateName=template, bdName=bd), l3Outs=[l3out.get("name")], l3OutRefs=[l3out_ref])
        else:
            mso_objects.get("site_bd").details["bdRef"] = dict(schemaId=mso_schema.id, templateName=template, bdName=bd)
            l3out_refs = mso_objects.get("site_bd").details.get("l3OutRefs", [])
            l3outs = mso_objects.get("site_bd").details.get("l3Outs", [])
            # check on name because refs are handled differently between versions
            if l3out.get("name") in l3outs:
                mso.existing = mso.dict_from_ref(l3out_refs[l3outs.index(l3out.get("name"))])

    if state == "query":
        if l3out is None:
            if "l3OutRefs" in mso_objects.get("site_bd", {}).details.keys():
                mso.existing = [mso.dict_from_ref(l3) for l3 in mso_objects.get("site_bd", {}).details.get("l3OutRefs", [])]
            else:
                mso.existing = [dict(l3outName=l3) for l3 in mso_objects.get("site_bd", {}).details.get("l3Outs", [])]
        elif not mso.existing:
            mso.fail_json(msg="L3out '{0}' not found".format(l3out.get("name")))
        mso.exit_json()

    mso.previous = mso.existing
    if state == "absent":
        if mso.existing:
            mso.sent = mso.existing = {}
            if l3out.get("name") in l3outs:
                del l3outs[l3outs.index(l3out.get("name"))]
            if l3out_ref in l3out_refs:
                del l3out_refs[l3out_refs.index(l3out_ref)]
            ops.append(dict(op="replace", path="{0}/{1}".format(bd_path, bd), value=mso_objects.get("site_bd").details))

    elif state == "present":
        if payload:
            ops.append(dict(op="add", path="{0}/-".format(bd_path), value=payload))
        elif not mso.existing:
            l3outs.append(l3out.get("name"))
            l3out_refs.append(l3out_ref)
            ops.append(dict(op="replace", path="{0}/{1}".format(bd_path, bd), value=mso_objects.get("site_bd").details))

        if l3out.get("tenant"):
            mso.existing = mso.dict_from_ref(l3out_ref)
        else:
            mso.existing = mso.make_reference(l3out, "l3out", l3out_schema_id, l3out_template)

    if not module.check_mode:
        mso.request(mso_schema.path, method="PATCH", data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2015, Jon Hawkesworth (@jhawkesworth) <figs@unity.demon.co.uk>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r'''
---
module: win_file
short_description: Creates, touches or removes files or directories
description:
     - Creates (empty) files, updates file modification stamps of existing files,
       and can create or remove directories.
     - Timestamp values are interpreted as local time on the target Windows system if no time zone offset is specified.
     - Unlike M(ansible.builtin.file), does not modify ownership, permissions or manipulate links.
     - For non-Windows targets, use the M(ansible.builtin.file) module instead.
options:
  path:
    description:
      - Path to the file being managed.
    required: yes
    type: path
    aliases: [ dest, name ]
  state:
    description:
      - If C(directory), all immediate subdirectories will be created if they
        do not exist.
      - If C(file), the file will NOT be created if it does not exist, see the M(ansible.windows.win_copy)
        or M(ansible.windows.win_template) module if you want that behavior.
      - If C(absent), directories will be recursively deleted, and files will be removed.
      - If C(touch), an empty file will be created if the C(path) does not
        exist, while an existing file or directory will receive updated file access and
        modification times (similar to the way C(touch) works from the command line).
    type: str
    choices: [ absent, directory, file, touch ]
  modification_time:
    description:
      - The desired modification time for the file or directory.
      - A DateTime string in the format specified by O(modification_time_format).
      - The timestamp is interpreted as local time on the target system.
      - Timezone offsets are supported when included in the timestamp format.
        (for example, using C(z), C(zz) or C(zzz) format specifiers).
      - When unset, the default is V(preserve) when O(state=[file, directory]) and V(now) when O(state=touch).
    type: str
    version_added: 3.4.0
  modification_time_format:
    description:
      - The format to use when parsing C(modification_time).
      - Defaults to C(yyyy-MM-dd HH:mm:ss).
      - See L(.NET DateTime format strings,https://learn.microsoft.com/en-us/dotnet/standard/base-types/standard-date-and-time-format-strings).
    type: str
    default: yyyy-MM-dd HH:mm:ss
    version_added: 3.4.0
  access_time:
    description:
      - The desired access time for the file or directory.
      - A DateTime string in the format specified by O(access_time_format).
      - The timestamp is interpreted as local time on the target system.
      - Timezone offsets are supported when included in the timestamp format.
        (for example, using C(z), C(zz) or C(zzz) format specifiers).
      - When unset, the default is V(preserve) when O(state=[file, directory]) and V(now) when O(state=touch).
    type: str
    version_added: 3.4.0
  access_time_format:
    description:
      - The format to use when parsing C(access_time).
      - Defaults to C(yyyy-MM-dd HH:mm:ss).
      - See L(.NET DateTime format strings,https://learn.microsoft.com/en-us/dotnet/standard/base-types/standard-date-and-time-format-strings).
    type: str
    default: yyyy-MM-dd HH:mm:ss
    version_added: 3.4.0
seealso:
- module: ansible.builtin.file
- module: ansible.windows.win_acl
- module: ansible.windows.win_acl_inheritance
- module: ansible.windows.win_owner
- module: ansible.windows.win_stat
author:
- Jon Hawkesworth (@jhawkesworth)
- Bishal Prasad (@bishalprasad321)
'''

EXAMPLES = r'''
- name: Touch a file (creates if not present, updates modification time if present)
  ansible.windows.win_file:
    path: C:\Temp\foo.conf
    state: touch

- name: Remove a file, if present
  ansible.windows.win_file:
    path: C:\Temp\foo.conf
    state: absent

- name: Create directory structure
  ansible.windows.win_file:
    path: C:\Temp\folder\subfolder
    state: directory

- name: Remove directory structure
  ansible.windows.win_file:
    path: C:\Temp
    state: absent

- name: Touch a file and set modification and access times to now
  ansible.windows.win_file:
    path: C:\Temp\foo.conf
    state: touch
    modification_time: now
    access_time: now

- name: Set specific modification and access times for a file
  ansible.windows.win_file:
    path: C:\Temp\foo.conf
    state: touch
    modification_time: "2025-12-29 12:34:56"
    access_time: "2025-12-29 12:34:56"

- name: Set specific modification as UTC datetime
  ansible.windows.win_file:
    path: C:\Temp\foo.conf
    state: touch
    modification_time: "2025-12-29 12:34:56 +0"
    modification_time_format: "yyyy-MM-dd HH:mm:ss z"

- name: Create a directory and set the timestamps to now
  ansible.windows.win_file:
    path: C:\Temp\folder
    state: directory
    modification_time: now
    access_time: now
'''



# -*- coding: utf-8 -*-
# Copyright 2022 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The ios l2_interfaces fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.argspec.l2_interfaces.l2_interfaces import (
    L2_interfacesArgs,
)
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.rm_templates.l2_interfaces import (
    L2_interfacesTemplate,
)


class L2_interfacesFacts(object):
    """The ios l2_interfaces facts class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = L2_interfacesArgs.argument_spec

    def get_l2_interfaces_data(self, connection):
        return connection.get("show running-config | section ^interface")

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for L2_interfaces network resource

        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected conf

        :rtype: dictionary
        :returns: facts
        """
        facts = {}
        objs = []

        if not data:
            data = self.get_l2_interfaces_data(connection)

        # parse native config using the L2_interfaces template
        l2_interfaces_parser = L2_interfacesTemplate(lines=data.splitlines(), module=self._module)
        objs = list(l2_interfaces_parser.parse().values())

        def process_mode(obj):
            mode = ""
            if obj == "dot1q-tunnel":
                mode = "dot1q_tunnel"
            elif obj == "dynamic auto":
                mode = "dynamic_auto"
            elif obj == "dynamic desirable":
                mode = "dynamic_desirable"
            elif obj == "private-vlan host":
                mode = "private_vlan_host"
            elif obj == "private-vlan promiscuous":
                mode = "private_vlan_promiscuous"
            elif obj == "private-vlan trunk secondary":
                mode = "private_vlan_trunk"
            return mode

        def process_vlans(obj, vlan_type):
            vlans = []
            _vlans = obj.get("trunk")
            if _vlans.get(vlan_type):
                vlans.extend(_vlans.get(vlan_type))
            if _vlans.get(vlan_type + "_add"):
                for vlan_grp in _vlans.get(vlan_type + "_add"):
                    vlans.extend(vlan_grp)
                del _vlans[vlan_type + "_add"]
            return vlans

        def process_xconnect(obj):
            """
            Normalize xconnect dictionary.
            Since rm_templates now generates 'address' and 'vcid' directly,
            we map them straight to the argspec keys.
            """
            xconnect = obj.get("xconnect")
            if xconnect:
                normalized = {}
                if xconnect.get("address"):
                    normalized["address"] = xconnect.get("address")
                elif xconnect.get("peer"):
                    normalized["address"] = xconnect.get("peer")
                if xconnect.get("vcid") is not None:
                    normalized["vcid"] = xconnect.get("vcid")
                elif xconnect.get("vc_id") is not None:
                    normalized["vcid"] = xconnect.get("vc_id")
                if xconnect.get("encapsulation"):
                    normalized["encapsulation"] = xconnect.get("encapsulation")
                if xconnect.get("pw_class"):
                    normalized["pw_class"] = xconnect.get("pw_class")
                if xconnect.get("sequencing"):
                    normalized["sequencing"] = xconnect.get("sequencing")
                if xconnect.get("manual") is not None:
                    normalized["manual"] = xconnect.get("manual")
                return normalized if normalized else None
            return None

        if objs:
            for obj in objs:
                if obj.get("mode") and obj.get("mode") not in ["trunk", "access", "dynamic"]:
                    obj["mode"] = process_mode(obj.get("mode"))
                if obj.get("trunk"):
                    for _vlan in ["allowed_vlans", "pruning_vlans"]:
                        if obj.get("trunk", {}).get(_vlan):
                            obj["trunk"][_vlan] = process_vlans(obj, _vlan)
                if obj.get("xconnect"):
                    obj["xconnect"] = process_xconnect(obj)

            ansible_facts["ansible_network_resources"].pop("l2_interfaces", None)

        params = utils.remove_empties(
            l2_interfaces_parser.validate_config(self.argument_spec, {"config": objs}, redact=True),
        )

        facts["l2_interfaces"] = params.get("config", [])
        ansible_facts["ansible_network_resources"].update(facts)

        return ansible_facts



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_early_access_features_opt_ins
short_description: Resource module for organizations _early _access _features _opt
  _ins
description:
  - Manage operations create, update and delete of the resource organizations _early
    _access _features _opt _ins.
  - Create a new early access feature opt-in for an organization.
  - Delete an early access feature opt-in.
  - Update an early access feature opt-in for an organization.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  limitScopeToNetworks:
    description: A list of network IDs to apply the opt-in to.
    elements: str
    type: list
  optInId:
    description: OptInId path parameter. Opt in ID.
    type: str
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
  shortName:
    description: Short name of the early access feature.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations createOrganizationEarlyAccessFeaturesOptIn
    description: Complete reference of the createOrganizationEarlyAccessFeaturesOptIn
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-organization-early-access-features-opt-in
  - name: Cisco Meraki documentation for organizations deleteOrganizationEarlyAccessFeaturesOptIn
    description: Complete reference of the deleteOrganizationEarlyAccessFeaturesOptIn
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!delete-organization-early-access-features-opt-in
  - name: Cisco Meraki documentation for organizations updateOrganizationEarlyAccessFeaturesOptIn
    description: Complete reference of the updateOrganizationEarlyAccessFeaturesOptIn
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-organization-early-access-features-opt-in
notes:
  - SDK Method used are
    organizations.Organizations.create_organization_early_access_features_opt_in,
    organizations.Organizations.delete_organization_early_access_features_opt_in,
    organizations.Organizations.update_organization_early_access_features_opt_in,
  - Paths used are
    post /organizations/{organizationId}/earlyAccess/features/optIns,
    delete /organizations/{organizationId}/earlyAccess/features/optIns/{optInId},
    put /organizations/{organizationId}/earlyAccess/features/optIns/{optInId},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_early_access_features_opt_ins:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    limitScopeToNetworks:
      - N_12345
    organizationId: string
    shortName: has_beta_api
- name: Update by id
  cisco.meraki.organizations_early_access_features_opt_ins:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    limitScopeToNetworks:
      - N_12345
    optInId: string
    organizationId: string
- name: Delete by id
  cisco.meraki.organizations_early_access_features_opt_ins:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: absent
    optInId: string
    organizationId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "id": "string",
      "shortName": "string",
      "limitScopeToNetworks": [
        {
          "id": "string",
          "name": "string"
        }
      ],
      "optOutEligibility": {
        "eligible": true,
        "reason": "string",
        "help": {
          "label": "string",
          "url": "string"
        }
      },
      "createdAt": "string"
    }
"""



# Copyright (c) 2025, Max Mitschke <maxmitschke@fastmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: print_task
type: aggregate
short_description: Prints playbook task snippet to job output
description:
  - This plugin prints the currently executing playbook task to the job output.
version_added: 10.7.0
requirements:
  - enable in configuration
"""

EXAMPLES = r"""
ansible.cfg: |-
  # Enable plugin
  [defaults]
  callbacks_enabled=community.general.print_task
"""

from yaml import dump, load

try:
    from yaml import CSafeDumper as SafeDumper
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeDumper, SafeLoader  # type: ignore

from ansible.plugins.callback import CallbackBase


class CallbackModule(CallbackBase):
    """
    This callback module tells you how long your plays ran for.
    """

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = "aggregate"
    CALLBACK_NAME = "community.general.print_task"

    CALLBACK_NEEDS_ENABLED = True

    def __init__(self):
        super().__init__()
        self._printed_message = False

    def _print_task(self, task):
        if hasattr(task, "_ds"):
            task_snippet = load(str([task._ds.copy()]), Loader=SafeLoader)
            task_yaml = dump(task_snippet, sort_keys=False, Dumper=SafeDumper)
            self._display.display(f"\n{task_yaml}\n")
            self._printed_message = True

    def v2_playbook_on_task_start(self, task, is_conditional):
        self._printed_message = False

    def v2_runner_on_start(self, host, task):
        if not self._printed_message:
            self._print_task(task)



#!/usr/bin/python
#
# Copyright (c) 2023, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: docker_image_export

short_description: Export (archive) Docker images

version_added: 3.7.0

description:
  - Creates an archive (tarball) from one or more Docker images.
  - This can be copied to another machine and loaded with M(community.docker.docker_image_load).
extends_documentation_fragment:
  - community.docker._docker.api_documentation
  - community.docker._attributes
  - community.docker._attributes.actiongroup_docker

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
  idempotent:
    support: partial
    details:
      - Whether the module is idempotent depends on the storage API used for images,
        which determines how the image ID is computed. The idempotency check needs
        the image ID to equal the ID stored in archive's C(manifest.json).
        This seemed to have worked fine with the default storage backend up to Docker 28,
        but seems to have changed in Docker 29.
      - This module is B(not idempotent) when used with multi-architecture images,
        regardless of Docker version.
      - Full idempotency requires Docker 28 or earlier B(and) a single-architecture image.

options:
  names:
    description:
      - 'One or more image names. Name format will be one of: C(name), C(repository/name), C(registry_server:port/name). When
        pushing or pulling an image the name can optionally include the tag by appending C(:tag_name).'
      - Note that image IDs (hashes) can also be used.
    type: list
    elements: str
    required: true
    aliases:
      - name
  tag:
    description:
      - Tag for the image name O(name) that is to be tagged.
      - If O(name)'s format is C(name:tag), then the tag value from O(name) will take precedence.
    type: str
    default: latest
  path:
    description:
      - The C(.tar) file the image should be exported to.
    type: path
  force:
    description:
      - Export the image even if the C(.tar) file already exists and seems to contain the right image.
    type: bool
    default: false
  platform:
    description:
      - Ask for this specific platform when exporting.
      - For example, C(linux/amd64), C(linux/arm64).
      - Requires Docker API 1.48 or newer.
    type: str
    version_added: 5.2.0

requirements:
  - "Docker API >= 1.25"

author:
  - Felix Fontein (@felixfontein)

seealso:
  - module: community.docker.docker_image
  - module: community.docker.docker_image_info
  - module: community.docker.docker_image_load
"""

EXAMPLES = r"""
---
- name: Export an image
  community.docker.docker_image_export:
    name: pacur/centos-7
    path: /tmp/centos-7.tar

- name: Export multiple images
  community.docker.docker_image_export:
    names:
      - hello-world:latest
      - pacur/centos-7:latest
    path: /tmp/various.tar
"""

RETURN = r"""
images:
  description: Image inspection results for the affected images.
  returned: success
  type: list
  elements: dict
  sample: []
"""

import json
import traceback
import typing as t

from ansible_collections.community.docker.plugins.module_utils._api.constants import (
    DEFAULT_DATA_CHUNK_SIZE,
)
from ansible_collections.community.docker.plugins.module_utils._api.errors import (
    DockerException,
)
from ansible_collections.community.docker.plugins.module_utils._api.utils.utils import (
    parse_repository_tag,
)
from ansible_collections.community.docker.plugins.module_utils._common_api import (
    AnsibleDockerClient,
    RequestException,
)
from ansible_collections.community.docker.plugins.module_utils._image_archive import (
    ImageArchiveInvalidException,
    api_image_id,
    load_archived_image_manifest,
)
from ansible_collections.community.docker.plugins.module_utils._platform import (
    _Platform,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    DockerBaseClass,
    is_image_name_id,
    is_valid_tag,
)


class ImageExportManager(DockerBaseClass):
    def __init__(self, client: AnsibleDockerClient) -> None:
        super().__init__()

        self.client = client
        parameters = self.client.module.params
        self.check_mode = self.client.check_mode

        self.path = parameters["path"]
        self.force = parameters["force"]
        self.tag = parameters["tag"]
        self.platform = parameters["platform"]

        if not is_valid_tag(self.tag, allow_empty=True):
            self.fail(f'"{self.tag}" is not a valid docker tag')

        # If name contains a tag, it takes precedence over tag parameter.
        self.names = []
        for name in parameters["names"]:
            if is_image_name_id(name):
                self.names.append({"id": name, "joined": name})
            else:
                repo, repo_tag = parse_repository_tag(name)
                if not repo_tag:
                    repo_tag = self.tag
                self.names.append(
                    {"name": repo, "tag": repo_tag, "joined": f"{repo}:{repo_tag}"}
                )

        if not self.names:
            self.fail("At least one image name must be specified")

    def fail(self, msg: str) -> t.NoReturn:
        self.client.fail(msg)

    def get_export_reason(self) -> str | None:
        if self.force:
            return "Exporting since force=true"

        try:
            archived_images = load_archived_image_manifest(self.path)
            if archived_images is None:
                return "Overwriting since no image is present in archive"
        except ImageArchiveInvalidException as exc:
            self.log(f"Unable to extract manifest summary from archive: {exc}")
            return "Overwriting an unreadable archive file"

        left_names = list(self.names)
        for archived_image in archived_images:
            found = False
            for i, name in enumerate(left_names):
                if (
                    name["id"] == api_image_id(archived_image.image_id)
                    and [name["joined"]] == archived_image.repo_tags
                ):
                    del left_names[i]
                    found = True
                    break
            if not found:
                return f"Overwriting archive since it contains unexpected image {archived_image.image_id} named {', '.join(archived_image.repo_tags)}"
        if left_names:
            return f"Overwriting archive since it is missing image(s) {', '.join([name['joined'] for name in left_names])}"

        return None

    def write_chunks(self, chunks: t.Generator[bytes]) -> None:
        try:
            with open(self.path, "wb") as fd:
                for chunk in chunks:
                    fd.write(chunk)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(f"Error writing image archive {self.path} - {exc}")

    def _platform_param(self) -> str:
        platform = _Platform.parse_platform_string(self.platform)
        platform_spec: dict[str, str] = {}
        if platform.os:
            platform_spec["os"] = platform.os
        if platform.arch:
            platform_spec["architecture"] = platform.arch
        if platform.variant:
            platform_spec["variant"] = platform.variant
        return json.dumps(platform_spec)

    def export_images(self) -> None:
        image_names = [name["joined"] for name in self.names]
        image_names_str = ", ".join(image_names)
        if len(image_names) == 1:
            self.log(f"Getting archive of image {image_names[0]}")
            params: dict[str, t.Any] = {}
            if self.platform:
                params["platform"] = self._platform_param()
            try:
                chunks = self.client._stream_raw_result(
                    self.client._get(
                        self.client._url("/images/{0}/get", image_names[0]),
                        stream=True,
                        params=params,
                    ),
                    chunk_size=DEFAULT_DATA_CHUNK_SIZE,
                    decode=False,
                )
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error getting image {image_names[0]} - {exc}")
        else:
            self.log(f"Getting archive of images {image_names_str}")
            params = {"names": image_names}
            if self.platform:
                params["platform"] = self._platform_param()
            try:
                chunks = self.client._stream_raw_result(
                    self.client._get(
                        self.client._url("/images/get"),
                        stream=True,
                        params=params,
                    ),
                    chunk_size=DEFAULT_DATA_CHUNK_SIZE,
                    decode=False,
                )
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error getting images {image_names_str} - {exc}")

        self.write_chunks(chunks)

    def run(self) -> dict[str, t.Any]:
        tag = self.tag
        if not tag:
            tag = "latest"

        images = []
        for name in self.names:
            if "id" in name:
                image = self.client.find_image_by_id(
                    name["id"], accept_missing_image=True
                )
            else:
                image = self.client.find_image(name=name["name"], tag=name["tag"])
            if not image:
                self.fail(f"Image {name['joined']} not found")
            images.append(image)

            # Will have a 'sha256:' prefix
            name["id"] = image["Id"]

        results = {
            "changed": False,
            "images": images,
        }

        reason = self.get_export_reason()
        if reason is not None:
            results["msg"] = reason
            results["changed"] = True

            if not self.check_mode:
                self.export_images()

        return results


def main() -> None:
    argument_spec = {
        "path": {"type": "path"},
        "force": {"type": "bool", "default": False},
        "names": {
            "type": "list",
            "elements": "str",
            "required": True,
            "aliases": ["name"],
        },
        "tag": {"type": "str", "default": "latest"},
        "platform": {"type": "str"},
    }

    option_minimal_versions = {
        "platform": {"docker_api_version": "1.48"},
    }

    client = AnsibleDockerClient(
        argument_spec=argument_spec,
        supports_check_mode=True,
        option_minimal_versions=option_minimal_versions,
    )

    try:
        results = ImageExportManager(client).run()
        client.module.exit_json(**results)
    except DockerException as e:
        client.fail(
            f"An unexpected Docker error occurred: {e}",
            exception=traceback.format_exc(),
        )
    except RequestException as e:
        client.fail(
            f"An unexpected requests error occurred when trying to talk to the Docker daemon: {e}",
            exception=traceback.format_exc(),
        )


if __name__ == "__main__":
    main()



# Copyright (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: pfexec
short_description: Profile based execution
description:
  - This become plugins allows your remote/login user to execute commands as another user using the C(pfexec) utility.
author: Ansible Core Team
options:
  become_user:
    description:
      - User you 'become' to execute the task.
      - This plugin ignores this setting as pfexec uses its own C(exec_attr) to figure this out, but it is supplied here for
        Ansible to make decisions needed for the task execution, like file permissions.
    type: string
    default: root
    ini:
      - section: privilege_escalation
        key: become_user
      - section: pfexec_become_plugin
        key: user
    vars:
      - name: ansible_become_user
      - name: ansible_pfexec_user
    env:
      - name: ANSIBLE_BECOME_USER
      - name: ANSIBLE_PFEXEC_USER
  become_exe:
    description: C(pfexec) executable.
    type: string
    default: pfexec
    ini:
      - section: privilege_escalation
        key: become_exe
      - section: pfexec_become_plugin
        key: executable
    vars:
      - name: ansible_become_exe
      - name: ansible_pfexec_exe
    env:
      - name: ANSIBLE_BECOME_EXE
      - name: ANSIBLE_PFEXEC_EXE
  become_flags:
    description: Options to pass to C(pfexec).
    type: string
    default: ""
    ini:
      - section: privilege_escalation
        key: become_flags
      - section: pfexec_become_plugin
        key: flags
    vars:
      - name: ansible_become_flags
      - name: ansible_pfexec_flags
    env:
      - name: ANSIBLE_BECOME_FLAGS
      - name: ANSIBLE_PFEXEC_FLAGS
  become_pass:
    description: C(pfexec) password.
    type: string
    required: false
    vars:
      - name: ansible_become_password
      - name: ansible_become_pass
      - name: ansible_pfexec_pass
    env:
      - name: ANSIBLE_BECOME_PASS
      - name: ANSIBLE_PFEXEC_PASS
    ini:
      - section: pfexec_become_plugin
        key: password
  wrap_exe:
    description:
      - Toggle to wrap the command C(pfexec) calls in C(shell -c) or not.
      - Unlike C(sudo), C(pfexec) does not interpret shell constructs internally,
        so commands containing shell operators must be wrapped in a shell invocation.
      - The current default of V(false) only works in very limited cases (for example
        with M(ansible.builtin.raw)).
      - The current default is B(deprecated) and will change to V(true) in community.general 14.0.0.
        To avoid the deprecation message, you can explicitly set this option to a value.
    type: bool
    ini:
      - section: pfexec_become_plugin
        key: wrap_execution
    vars:
      - name: ansible_pfexec_wrap_execution
    env:
      - name: ANSIBLE_PFEXEC_WRAP_EXECUTION
notes:
  - This plugin ignores O(become_user) as pfexec uses its own C(exec_attr) to figure this out.
"""

from ansible.plugins.become import BecomeBase
from ansible.utils.display import Display

display = Display()


class BecomeModule(BecomeBase):
    name = "community.general.pfexec"

    def build_become_command(self, cmd, shell):
        super().build_become_command(cmd, shell)

        if not cmd:
            return cmd

        exe = self.get_option("become_exe")
        flags = self.get_option("become_flags")

        wrap_exe = self.get_option("wrap_exe")
        if wrap_exe is None:
            display.deprecated(
                "The default value of the wrap_exe option for the community.general.pfexec "
                "become plugin will change from false to true in community.general 14.0.0. "
                "Set wrap_exe explicitly to silence this warning.",
                version="14.0.0",
                collection_name="community.general",
            )
            wrap_exe = False

        become_cmd = self._build_success_command(cmd, shell, noexe=not wrap_exe)
        return " ".join(part for part in (exe, flags, become_cmd) if part)



#
# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The eos_ntp_global config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.arista.eos.plugins.module_utils.network.eos.facts.facts import Facts
from ansible_collections.arista.eos.plugins.module_utils.network.eos.rm_templates.ntp_global import (
    Ntp_globalTemplate,
)
from ansible_collections.arista.eos.plugins.module_utils.network.eos.utils.utils import (
    normalize_interface,
)


class Ntp_global(ResourceModule):
    """
    The eos_ntp_global config class
    """

    def __init__(self, module):
        super(Ntp_global, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="ntp_global",
            tmplt=Ntp_globalTemplate(),
        )
        self.parsers = [
            "authenticate",
            "local_interface",
            "qos_dscp",
            "trusted_key",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """
        wantd = {"ntp_global": self.want}
        haved = {"ntp_global": self.have}
        # turn all lists of dicts into dicts prior to merge
        for entry in wantd["ntp_global"], haved["ntp_global"]:
            self._ntp_global_list_to_dict(entry)

        # if state is merged, merge want onto have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, empty out wantd and set haved to wantd
        if self.state == "deleted":
            wantd = {}
            for k, have in haved.items():
                if k not in wantd:
                    self._compare(want={}, have=have)

        for k, want in wantd.items():
            self._compare(want=want, have=haved.pop(k, {}))

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Ntp_global network resource.
        """
        self._serve_compare(want=want, have=have)
        self._authentication_keys_compare(want=want, have=have)
        self._servers_compare(want=want, have=have)
        self.compare(parsers=self.parsers, want=want, have=have)
        add_cmd = []
        del_cmd = []
        if self.commands:
            for cmd in self.commands:
                if "no ntp" in cmd:
                    del_cmd.append(cmd)
                else:
                    add_cmd.append(cmd)
            self.commands = del_cmd + add_cmd

    def _authentication_keys_compare(self, want, have):
        w = want.pop("authentication_keys", {})
        h = have.pop("authentication_keys", {})
        for name, entry in w.items():
            h_key = {}
            if h.get(name):
                h_key = {"authentication_keys": h.pop(name)}
            self.compare(
                parsers="authentication_keys",
                want={"authentication_keys": entry},
                have=h_key,
            )
        for name, entry in h.items():
            self.compare(
                parsers="authentication_keys",
                want={},
                have={"authentication_keys": entry},
            )

    def _servers_compare(self, want, have):
        w = want.pop("servers", {})
        h = have.pop("servers", {})
        for name, entry in w.items():
            if entry.get("source"):
                entry["source"] = normalize_interface(entry["source"])
            h_key = {}
            if h.get(name):
                h_key = {"servers": h.pop(name)}
            self.compare(
                parsers="servers",
                want={"servers": entry},
                have=h_key,
            )
        for name, entry in h.items():
            self.compare(parsers="servers", want={}, have={"servers": entry})

    def _serve_compare(self, want, have):
        serve_want = want.pop("serve", {})
        serve_have = have.pop("serve", {})
        for name, entry in serve_want.items():
            if name == "all" and entry:
                w = {"serve": {"all": True}}
                self.compare(
                    parsers="serve_all",
                    want=w,
                    have={"serve": {"all": serve_have.pop("all", False)}},
                )
            else:
                for k_afi, v_afi in entry.items():
                    for k, v in v_afi.items():
                        afi = v_afi["afi"]
                        if k == "afi":
                            continue
                        h = {}
                        if k == "acls":
                            for ace, ace_entry in v.items():
                                if serve_have.get("access_lists"):
                                    for hk, hv in serve_have["access_lists"].items():
                                        for h_k, h_v in hv.items():
                                            h_afi = hv["afi"]
                                            if h_k == "afi":
                                                continue
                                            if h_afi == afi:
                                                if ace in h_v:
                                                    h_acc = {
                                                        "afi": h_afi,
                                                        "acls": h_v.pop(ace),
                                                    }
                                                    h = {
                                                        "serve": {
                                                            "access_lists": h_acc,
                                                        },
                                                    }
                                w = {
                                    "serve": {
                                        "access_lists": {
                                            "afi": afi,
                                            "acls": ace_entry,
                                        },
                                    },
                                }
                                self.compare(parsers="serve", want=w, have=h)
        for k, v in serve_have.items():
            if k == "all" and v:
                h = {"serve": {"all": True}}
                self.compare(parsers="serve_all", want={}, have=h)
            else:
                for k_afi, v_afi in v.items():
                    for k, v in v_afi.items():
                        hafi = v_afi["afi"]
                        if k == "afi":
                            continue
                        for k_acl, v_acl in v.items():
                            h = {
                                "serve": {
                                    "access_lists": {
                                        "afi": hafi,
                                        "acls": v_acl,
                                    },
                                },
                            }
                            self.compare(parsers="serve", want={}, have=h)

    def _ntp_global_list_to_dict(self, entry):
        if "authentication_keys" in entry:
            key_dict = {}
            for el in entry["authentication_keys"]:
                key_dict.update({el["id"]: el})
            entry["authentication_keys"] = key_dict

        if "servers" in entry:
            server_dict = {}
            for el in entry["servers"]:
                server_dict.update({el["server"]: el})
            entry["servers"] = server_dict

        if "serve" in entry:
            serve_dict = {}
            main_dict = {}
            if entry["serve"].get("all"):
                main_dict.update({"all": entry["serve"]["all"]})
            if entry["serve"].get("access_lists"):
                for el in entry["serve"].get("access_lists"):
                    if "acls" in el:
                        acl_dict = {}
                        for acl in el["acls"]:
                            acl_dict.update({acl["acl_name"]: acl})
                        el["acls"] = acl_dict
                    serve_dict.update({el["afi"]: el})
                if serve_dict:
                    main_dict.update({"access_lists": serve_dict})
            if serve_dict:
                entry["serve"] = main_dict
